"""Timeline theme data structures."""

import os
from typing import Optional

from qt_api import QColor, QPixmap, QByteArray
from classes.info import PATH


def _apply_overrides(obj, overrides: dict, *, allow_unknown: bool = False) -> None:
    if not overrides:
        return
    allowed = set(obj.__dict__.keys())
    for key, value in overrides.items():
        if key not in allowed:
            if allow_unknown:
                continue
            raise TypeError("Unexpected theme option '%s'" % key)
        setattr(obj, key, value)


def _icon(rel_path: Optional[str]) -> Optional[QPixmap]:
    """Load a pixmap from *rel_path* (relative to PATH) with SVG metadata attached."""
    if not rel_path:
        return None
    path = os.path.normpath(os.path.join(PATH, rel_path))
    if not os.path.exists(path):
        return None
    pix = QPixmap(path)
    if pix.isNull():
        return None
    pix.svg_path = path
    if path.lower().endswith(".svg"):
        try:
            with open(path, "rb") as fh:
                data = fh.read()
            pix.svg_bytes = data
            try:
                pix.svg_qbytearray = QByteArray(data)
            except Exception:
                pass
        except Exception:
            pass
    return pix


class BasicTheme:
    """Common style options for timeline elements."""

    def __init__(self, **kwargs):
        self.background: QColor = QColor()
        self.background2: QColor = QColor()
        self.border_color: QColor = QColor()
        self.border_radius: int = 0
        self.border_width: float = 0
        self.font_color: QColor = QColor()
        self.font_size: int = 0
        self.height: int = 0
        self.background_image: Optional[QPixmap] = None
        self.shadow_color: QColor = QColor()
        self.shadow_blur: int = 0
        self.thumb_width: int = 0
        self.thumb_height: int = 0
        self.top_overlay: QColor = QColor()
        self.top_overlay2: QColor = QColor()
        _apply_overrides(self, kwargs)


class TrackTheme(BasicTheme):
    """Theme for tracks."""

    def __init__(self, **kwargs):
        super().__init__()
        self.name_background: QColor = QColor()
        self.name_width: int = 0
        self.gap: int = 0
        self.margin_top: int = -1
        self.name_border_color: QColor = QColor()
        self.name_border_width: int = 0
        self.name_border_top_color: QColor = QColor()
        self.name_border_top_width: int = 0
        self.name_border_bottom_color: QColor = QColor()
        self.name_border_bottom_width: int = 0
        self.name_radius_tl: int = 0
        self.name_radius_bl: int = 0
        self.name_top_overlay: QColor = QColor()
        self.name_top_overlay2: QColor = QColor()
        _apply_overrides(self, kwargs)


class TimelineTheme:
    """Container for all timeline painting values."""

    def __init__(self, **kwargs):
        self.background: QColor = QColor("#000")
        self.background2: QColor = QColor()
        self.playhead_color: QColor = QColor("#FFF")
        self.playhead_width: float = 0.0
        self.clip_selected: QColor = QColor("#FFF")
        self.selection: QColor = QColor(255, 255, 255, 80)
        self.selection_border: QColor = QColor()
        self.selection_border_width: float = 0.0
        self.playback_cache_color: QColor = QColor("#4B92AD")
        self.playback_cache_height: float = 5.0

        self.clip: BasicTheme = BasicTheme()
        self.transition: BasicTheme = BasicTheme()
        self.track: TrackTheme = TrackTheme()
        self.ruler: BasicTheme = BasicTheme()
        self.ruler_name_background: QColor = QColor()
        self.ruler_name_background2: QColor = QColor()
        self.ruler_time_font_size: int = 0
        self.menu_icon: Optional[QPixmap] = None
        self.menu_size: int = 0
        self.menu_margin: int = 0
        self.keyframe_toggle_off_icon: Optional[QPixmap] = None
        self.keyframe_toggle_on_icon: Optional[QPixmap] = None
        self.track_keyframe_panel_disabled_icon: Optional[QPixmap] = None
        self.track_keyframe_panel_enabled_icon: Optional[QPixmap] = None
        self.track_add_above_disabled_icon: Optional[QPixmap] = None
        self.track_add_above_enabled_icon: Optional[QPixmap] = None
        self.track_add_below_disabled_icon: Optional[QPixmap] = None
        self.track_add_below_enabled_icon: Optional[QPixmap] = None
        self.track_delete_disabled_icon: Optional[QPixmap] = None
        self.track_delete_enabled_icon: Optional[QPixmap] = None
        self.track_locked_disabled_icon: Optional[QPixmap] = None
        self.track_locked_enabled_icon: Optional[QPixmap] = None
        self.track_unlocked_disabled_icon: Optional[QPixmap] = None
        self.track_unlocked_enabled_icon: Optional[QPixmap] = None
        self.keyframe_panel_add_icon: Optional[QPixmap] = None
        self.keyframe_panel_property_bg: QColor = QColor()
        self.keyframe_panel_row_border_color: QColor = QColor()
        self.keyframe_panel_row_border_width: float = 1.0
        self.keyframe_panel_curve_color: QColor = QColor()
        self.keyframe_panel_marker_fill: QColor = QColor()
        self.keyframe_panel_marker_border: QColor = QColor()
        self.playhead_icon: Optional[QPixmap] = None
        self.playhead_icon_width: int = 0
        self.playhead_icon_height: int = 0
        self.playhead_icon_offset_x: int = 0
        self.playhead_icon_offset_y: int = 0
        self.marker_icon: Optional[QPixmap] = None
        self.marker_icon_width: int = 0
        self.marker_icon_height: int = 0
        self.marker_icon_offset_x: Optional[int] = None
        self.marker_icon_offset_y: Optional[int] = None
        self.marker_hit_padding: float = 4.0
        self.ruler_time_pad_left: int = 0
        self.ruler_time_pad_top: int = 0
        self.ruler_label_top: int = 0
        self.scrollbar_handle: QColor = QColor()
        self.scrollbar_track: QColor = QColor()
        self.scrollbar_width: int = 0
        self.waveform_color: QColor = QColor(42, 130, 218)
        self.waveform_peak_color: QColor = QColor(42, 130, 218, 128)
        self.keyframe_fill: QColor = QColor("#4d7bff")
        self.keyframe_border: QColor = QColor("#ffffff")
        self.keyframe_inactive_opacity: float = 0.5
        self.keyframe_size: int = 10
        _apply_overrides(self, kwargs)


DEFAULT_THEME = TimelineTheme()



"""
 @file
 @brief This file contains optimized-preview proxy management logic.
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import copy
import json
import os
import re
import threading
import uuid
from concurrent.futures import ThreadPoolExecutor

from qt_api import QObject, pyqtSignal, pyqtSlot
from qt_api import QFileDialog

import openshot

from classes import info
from classes.app import get_app
from classes.assets import get_assets_path
from classes.logger import log
from classes.path_utils import absolute_media_path, comparable_media_path
from classes.query import Clip, File, Transition
from classes.thumbnail import (
    GenerateThumbnailFromFrame,
    RoundFrameToThumbnailGrid,
    ThumbnailPathForFrame,
)
from classes.updates import UpdateAction


def dialog_preview_reader_data(file_obj, prefer_proxy=True):
     """Return reader payload for dialog previews, optionally preferring a valid proxy reader."""
     data = getattr(file_obj, "data", {}) if file_obj else {}
     if not isinstance(data, dict):
         data = {}

     file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
     proxy_reader = data.get("proxy_reader")
     if prefer_proxy and isinstance(proxy_reader, dict):
         if proxy_reader.get("missing"):
             proxy_reader = None
         else:
             proxy_path = absolute_media_path(proxy_reader.get("path"))
             if proxy_path and os.path.exists(proxy_path):
                 resolved = copy.deepcopy(proxy_reader)
                 resolved["path"] = proxy_path
                 if file_id:
                     resolved["id"] = file_id
                 return resolved

     resolved = copy.deepcopy(data)
     source_path = absolute_media_path(resolved.get("path"))
     if source_path:
         resolved["path"] = source_path
     if file_id:
         resolved["id"] = file_id
     return resolved


class ProxyService(QObject):
     """Encapsulates optimized-preview proxy generation, linking, and runtime rewrites."""

     proxy_generated = pyqtSignal(str, object, str)
     file_job_changed = pyqtSignal(str)
     queue_changed = pyqtSignal()
     job_updated = pyqtSignal(str, str, int)
     job_finished = pyqtSignal(str, str)
     ACTIVE_STATES = ("queued", "running", "canceling")
     OPTIMIZE_CACHE_MAX_BYTES = 16 * 1024 * 1024
     OPTIMIZE_CACHE_CLEAR_INTERVAL = 120
     MATCHABLE_VIDEO_EXTENSIONS = {
         ".3g2", ".3gp", ".asf", ".avi", ".f4v", ".flv", ".m2t", ".m2ts", ".m4v", ".mkv",
         ".mov", ".mp2", ".mp4", ".mpeg", ".mpg", ".mts", ".mxf", ".ogv", ".rm", ".rmvb",
         ".ts", ".vob", ".webm", ".wmv",
     }
     _PROXY_PREFIX_RE = re.compile(
         r"^(?:proxy|proxies|prox|optimized?|optimised|preview|lowres|low_res|low-res|offline|transcoded?)[\s._-]+",
         re.IGNORECASE,
     )
     _PROXY_SUFFIX_RE = re.compile(
         r"[\s._-]+(?:proxy|proxies|prox|optimized?|optimised|preview|lowres|low_res|low-res|offline|transcoded?)$",
         re.IGNORECASE,
     )
     _RESOLUTION_SUFFIX_RE = re.compile(
         r"[\s._-]+(?:360p|540p|720p|1080p|1440p|2160p|4320p|4k|8k)$",
         re.IGNORECASE,
     )
     _CAMERA_PROXY_SUFFIX_RE = re.compile(r"[\s._-]+p$", re.IGNORECASE)

     def __init__(self, win):
         super().__init__(win)
         self.win = win
         self._executor = None
         self._jobs = {}
         self._lock = threading.RLock()
         self._ensure_executor()
         self.proxy_generated.connect(self._on_proxy_generated)

     def shutdown(self):
         if getattr(self, "_executor", None):
             self._shutdown_executor(self._executor)

     def create_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         self._ensure_executor()
         os.makedirs(self._proxy_root(), exist_ok=True)
         submitted = 0
         linked_existing = 0
         skipped = 0
         for file_obj in files:
             file_id = str(file_obj.id or "")
             if not file_id:
                 continue
             if self.get_active_job_for_file(file_id):
                 log.info("Optimize Preview create_for_files file_id=%s skipped: job already active", file_id)
                 skipped += 1
                 continue
             if self.has_proxy_reader(file_obj) and not self.has_missing_proxy(file_obj):
                 log.info("Optimize Preview create_for_files file_id=%s skipped: already optimized", file_id)
                 skipped += 1
                 continue
             snapshot = copy.deepcopy(file_obj.data or {})
             existing_output_path = self._existing_proxy_output_path(file_id, snapshot)
             if existing_output_path:
                 try:
                     proxy_reader = self._reader_json_for_path(existing_output_path, file_id)
                     self._save_proxy_reader(file_id, proxy_reader)
                     log.info("Optimize Preview create_for_files file_id=%s linked existing output=%s", file_id, existing_output_path)
                     linked_existing += 1
                     continue
                 except Exception as exc:
                     log.warning(
                         "Optimize Preview create_for_files file_id=%s ignored invalid existing output=%s (%s)",
                         file_id,
                         existing_output_path,
                         exc,
                     )
                     try:
                         os.remove(existing_output_path)
                     except Exception as cleanup_exc:
                         log.warning(
                             "Optimize Preview create_for_files file_id=%s failed to remove invalid existing output=%s (%s)",
                             file_id,
                             existing_output_path,
                             cleanup_exc,
                         )
             output_path = self._reserve_proxy_output_path(file_id, snapshot)
             with self._lock:
                 self._jobs[file_id] = {
                     "id": file_id,
                     "status": "queued",
                     "progress": 0,
                     "cancel_requested": False,
                     "output_path": output_path,
                 }
             future = self._executor.submit(self._build_proxy_reader, file_id, snapshot)
             with self._lock:
                 if file_id in self._jobs:
                     self._jobs[file_id]["future"] = future
             future.add_done_callback(lambda fut, fid=file_id: self._emit_proxy_result(fid, fut))
             self._emit_job_change(file_id)
             submitted += 1

         if submitted:
             status = "Optimize Preview: creating {} item(s)".format(submitted)
             if linked_existing:
                 status += ", linked {}".format(linked_existing)
             if skipped:
                 status += ", skipped {}".format(skipped)
             self._show_status(status)
         elif linked_existing:
             status = "Optimize Preview: linked {} item(s)".format(linked_existing)
             if skipped:
                 status += ", skipped {}".format(skipped)
             self._show_status(status, 3000)
         elif skipped:
             self._show_status("Optimize Preview: skipped {} item(s)".format(skipped), 3000)

     def use_existing_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         start_dir = self._proxy_root()
         if not start_dir:
             start_dir = info.HOME_PATH
         translator = getattr(self.win, "_tr", None) or getattr(get_app(), "_tr", lambda text: text)
         selected_folder = QFileDialog.getExistingDirectory(
             self.win,
             translator("Choose optimized preview folder"),
             start_dir,
         )
         if not selected_folder:
             return

         matched = 0
         missing = 0
         invalid = 0
         changed_file_ids = []
         folder_index = self._index_existing_optimized_files(selected_folder)
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 match_path = self._match_existing_optimized_path(file_obj, selected_folder, folder_index)
                 if match_path and os.path.exists(match_path):
                     try:
                         proxy_reader = self._reader_json_for_path(match_path, file_obj.id)
                         matched += 1
                         self._save_proxy_reader(file_obj.id, proxy_reader, apply_runtime=False, emit_job_change=False)
                         changed_file_ids.append(str(file_obj.id or ""))
                     except Exception as exc:
                         log.warning(
                             "Optimize Preview use_existing_for_files file_id=%s skipped invalid existing file=%s (%s)",
                             file_obj.id,
                             match_path,
                             exc,
                         )
                         invalid += 1
                 else:
                     proxy_reader = self._missing_proxy_reader(file_obj, selected_folder)
                     self._save_proxy_reader(file_obj.id, proxy_reader, apply_runtime=False, emit_job_change=False)
                     changed_file_ids.append(str(file_obj.id or ""))
                     missing += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         status = "Optimize Preview: linked {} item(s), missing {}".format(matched, missing)
         if invalid:
             status += ", invalid {}".format(invalid)
         self._show_status(status, 4000)

     def remove_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return

         removed = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 file_id = str(file_obj.id or "")
                 self.cancel_job(file_id)
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 if "proxy_reader" not in data:
                     continue
                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 removed += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         if removed:
             self._show_status("Optimize Preview: removed {} item(s)".format(removed), 3000)

     def delete_and_unlink_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return 0

         deleted = 0
         unlinked = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in files:
                 file_id = str(file_obj.id or "")
                 self.cancel_job(file_id)
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 proxy_reader = data.get("proxy_reader")
                 if not isinstance(proxy_reader, dict):
                     continue

                 proxy_path = absolute_media_path(proxy_reader.get("path"))
                 if proxy_path and os.path.exists(proxy_path):
                     proxy_abs = os.path.abspath(proxy_path)
                     try:
                         os.remove(proxy_abs)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", proxy_abs, exc_info=1)

                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 unlinked += 1
         finally:
             updates.transaction_id = previous_tid

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         status = "Optimize Preview: deleted {}, unlinked {}".format(deleted, unlinked)
         if unlinked:
             self._show_status(status, 4000)
         return deleted

     def has_internal_project_proxy_files(self):
         proxy_root = self._proxy_root()
         if not proxy_root or not os.path.isdir(proxy_root):
             return False
         for _, _, files in os.walk(proxy_root):
             if files:
                 return True
         return False

     def delete_internal_project_proxy_files(self):
         proxy_root = os.path.abspath(str(self._proxy_root() or ""))
         proxy_root_prefix = proxy_root + os.sep if proxy_root else ""
         if not proxy_root_prefix or not os.path.isdir(proxy_root):
             return 0

         deleted = 0
         unlinked = 0
         changed_file_ids = []
         updates = get_app().updates
         previous_tid = getattr(updates, "transaction_id", None)
         updates.transaction_id = str(uuid.uuid4())
         try:
             for file_obj in File.filter():
                 file_id = str(getattr(file_obj, "id", "") or "")
                 if not file_id:
                     continue
                 fresh_file = File.get(id=file_id)
                 data = getattr(fresh_file, "data", {}) or {}
                 proxy_reader = data.get("proxy_reader")
                 if not isinstance(proxy_reader, dict):
                     continue

                 proxy_path = absolute_media_path(proxy_reader.get("path"))
                 if not proxy_path:
                     continue
                 proxy_abs = os.path.abspath(proxy_path)
                 if not proxy_abs.startswith(proxy_root_prefix):
                     continue

                 self.cancel_job(file_id)
                 if os.path.exists(proxy_abs):
                     try:
                         os.remove(proxy_abs)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", proxy_abs, exc_info=1)

                 data.pop("proxy_reader", None)
                 fresh_file.data = data
                 fresh_file.save()

                 refreshed = File.get(id=file_id)
                 refreshed_data = getattr(refreshed, "data", {}) or {}
                 if "proxy_reader" in refreshed_data and refreshed and refreshed.key:
                     get_app().updates.delete(refreshed.key + ["proxy_reader"])
                 changed_file_ids.append(file_id)
                 unlinked += 1
         finally:
             updates.transaction_id = previous_tid

         try:
             for root, dirs, files in os.walk(proxy_root, topdown=False):
                 for name in files:
                     file_path = os.path.join(root, name)
                     try:
                         os.remove(file_path)
                         deleted += 1
                     except Exception:
                         log.warning("Optimize Preview delete failed for %s", file_path, exc_info=1)
                 for name in dirs:
                     dir_path = os.path.join(root, name)
                     try:
                         os.rmdir(dir_path)
                     except OSError:
                         pass
         except Exception:
             log.warning("Optimize Preview cleanup failed for %s", proxy_root, exc_info=1)

         self.apply_runtime_updates_for_files(changed_file_ids)
         for file_id in changed_file_ids:
             self._emit_job_change(file_id)

         if deleted or unlinked:
             self._show_status("Optimize Preview: deleted {}, unlinked {}".format(deleted, unlinked), 4000)
         return deleted

     def cancel_for_files(self, files):
         files = [f for f in (files or []) if getattr(f, "id", None)]
         if not files:
             return False

         canceled = 0
         for file_obj in files:
             if self.cancel_job(file_obj.id):
                 canceled += 1

         if canceled:
             self._show_status("Optimize Preview: canceled {} item(s)".format(canceled), 3000)
             return True
         return False

     def cancel_job(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return False

         with self._lock:
             job = self._jobs.get(file_id)
             if not job:
                 return False
             status = str(job.get("status") or "")
             future = job.get("future")
             if status == "queued" and future and future.cancel():
                 self._finalize_job(file_id, "canceled")
                 return True
             if status == "queued":
                 job["cancel_requested"] = True
                 job["status"] = "canceling"
             elif status == "running":
                 job["cancel_requested"] = True
                 job["status"] = "canceling"
             elif status == "canceling":
                 return True
             else:
                 return False

         self._emit_job_change(file_id)
         return True

     def rewrite_json_for_preview(self, payload):
         """Return preview-bound JSON/object with proxy readers substituted when available."""
         original_is_text = isinstance(payload, str)
         if original_is_text:
             try:
                 parsed = json.loads(payload)
             except Exception:
                 return payload
         else:
             parsed = copy.deepcopy(payload)

         proxy_lookup, path_lookup = self._proxy_lookup_from_payload(parsed)
         if not proxy_lookup:
             proxy_lookup, path_lookup = self._proxy_lookup_from_project()
         if not proxy_lookup:
             return payload

         changed = self._rewrite_payload_in_place(parsed, proxy_lookup, path_lookup)
         if not changed:
             return payload
         if original_is_text:
             return json.dumps(parsed)
         return parsed

     def apply_runtime_updates_for_file(self, file_id):
         """Push runtime-only clip/effect updates to libopenshot for one file's proxy state."""
         return self.apply_runtime_updates_for_files([file_id])

     def apply_runtime_updates_for_files(self, file_ids):
         """Push one combined runtime-only clip/effect diff for one or more file proxy changes."""
         normalized_ids = {str(file_id or "") for file_id in (file_ids or []) if str(file_id or "")}
         if not normalized_ids:
             return False

         timeline_sync = getattr(self.win, "timeline_sync", None)
         timeline = getattr(timeline_sync, "timeline", None) if timeline_sync else None
         if timeline is None:
             return False

         actions = []
         for clip in Clip.filter():
             if any(self._payload_references_file(clip.data, file_id) for file_id in normalized_ids):
                 actions.append(
                     UpdateAction(
                         type="update",
                         key=["clips", {"id": clip.id}],
                         values=copy.deepcopy(clip.data),
                     )
                 )
         for effect in Transition.filter():
             if any(self._payload_references_file(effect.data, file_id) for file_id in normalized_ids):
                 actions.append(
                     UpdateAction(
                         type="update",
                         key=["effects", {"id": effect.id}],
                         values=copy.deepcopy(effect.data),
                     )
                 )

         if not actions:
             return False

         payload = [json.loads(action.json()) for action in actions]
         timeline.ApplyJsonDiff(self.rewrite_json_for_preview(json.dumps(payload)))
         try:
             timeline.ClearAllCache(True)
         except Exception:
             log.debug("Optimize Preview cache clear failed", exc_info=1)
         if getattr(self.win, "refreshFrameSignal", None):
             self.win.refreshFrameSignal.emit()
         return True

     def has_proxy_reader(self, file_obj):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         proxy_reader = data.get("proxy_reader") if isinstance(data, dict) else None
         return isinstance(proxy_reader, dict) and bool(proxy_reader.get("path"))

     def has_missing_proxy(self, file_obj):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         proxy_reader = data.get("proxy_reader") if isinstance(data, dict) else None
         if not isinstance(proxy_reader, dict) or not proxy_reader.get("path"):
             return False
         if proxy_reader.get("missing"):
             return True
         return not os.path.exists(str(proxy_reader.get("path")))

     def get_active_job_for_file(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return None
         with self._lock:
             job = self._job_snapshot(self._jobs.get(file_id))
         if not job or job.get("status") not in self.ACTIVE_STATES:
             return None
         return job

     def get_file_badge(self, file_id):
         job = self.get_active_job_for_file(file_id)
         if not job:
             return None

         status = str(job.get("status") or "").strip().lower()
         progress = int(job.get("progress", 0))
         if status == "queued":
             label = "Queued"
         elif status == "running":
             label = "Creating {}%".format(progress)
         elif status == "canceling":
             label = "Canceling..."
         else:
             label = status.capitalize()
         return {
             "status": status,
             "progress": progress,
             "label": label,
             "job_id": str(job.get("id") or ""),
         }

     def get_proxy_state(self, file_obj):
         file_id = str(getattr(file_obj, "id", "") or "")
         active_job = self.get_active_job_for_file(file_id)
         if active_job:
             return str(active_job.get("status") or "queued")
         if self.has_proxy_reader(file_obj):
             return "missing" if self.has_missing_proxy(file_obj) else "ready"
         return "none"

     def _proxy_root(self):
         app = get_app()
         project = getattr(app, "project", None) if app else None
         current_filepath = getattr(project, "current_filepath", None) if project else None
         if current_filepath:
             current_abs = os.path.abspath(str(current_filepath))
             backup_abs = os.path.abspath(info.BACKUP_FILE)
             recovery_abs = os.path.abspath(info.RECOVERY_PATH) + os.sep
             if current_abs == backup_abs or current_abs.startswith(recovery_abs):
                 return info.PROXY_PATH
             return os.path.join(get_assets_path(current_filepath), "optimized")
         return info.PROXY_PATH

     def _emit_proxy_result(self, file_id, future):
         error_text = ""
         proxy_reader = {}
         try:
             if future.cancelled():
                 raise _ProxyJobCanceled()
             proxy_reader = future.result()
         except _ProxyJobCanceled:
             error_text = "__canceled__"
         except Exception as ex:
             error_text = str(ex)
             log.warning("Optimized Preview generation failed for %s: %s", file_id, error_text, exc_info=1)
         self.proxy_generated.emit(str(file_id or ""), proxy_reader, error_text)

     @pyqtSlot(str, object, str)
     def _on_proxy_generated(self, file_id, proxy_reader, error_text):
         file_id = str(file_id or "")
         if error_text == "__canceled__":
             self._finalize_job(file_id, "canceled")
             return
         if error_text:
             self._finalize_job(file_id, "failed")
             self._show_status("Optimize Preview: {}".format(error_text), 5000)
             return
         self._save_proxy_reader(file_id, proxy_reader)
         self._finalize_job(file_id, "completed")
         self._show_status("Optimize Preview: ready", 3000)

     def _save_proxy_reader(self, file_id, proxy_reader, apply_runtime=True, emit_job_change=True):
         file_obj = File.get(id=file_id)
         if not file_obj:
             return
         file_obj.data["proxy_reader"] = copy.deepcopy(proxy_reader or {})
         file_obj.save()
         if apply_runtime:
             self.apply_runtime_updates_for_file(file_id)
         if emit_job_change:
             self._emit_job_change(file_id)

     def _reader_json_for_path(self, file_path, file_id):
         clip = openshot.Clip(str(file_path))
         try:
             clip.Open()
             proxy_reader = json.loads(clip.Reader().Json())
             proxy_reader["id"] = str(file_id or "")
             return proxy_reader
         finally:
             try:
                 clip.Close()
             except Exception:
                 pass

     def _build_proxy_reader(self, file_id, file_data):
         source_path = absolute_media_path(file_data.get("path"))
         if not source_path or not os.path.exists(source_path):
             raise RuntimeError("source file not found")
         if str(file_data.get("media_type", "")).lower() != "video":
             raise RuntimeError("only video files are supported in V1")

         self._mark_running(file_id)
         output_path = self._reserved_or_computed_proxy_output_path(file_id, file_data)
         try:
             max_width, max_height = self._max_optimize_bounds()
             optimize_timeline = self._create_optimize_timeline(max_width, max_height)
             clip = openshot.Clip(source_path)
             clip.ParentTimeline(optimize_timeline)
             clip.Open()
             try:
                 reader = clip.Reader()
                 self._configure_optimize_caches(clip, reader)
                 source_reader = json.loads(reader.Json())

                 width = int(source_reader.get("width", 0) or 0)
                 height = int(source_reader.get("height", 0) or 0)
                 if width <= 0 or height <= 0:
                     raise RuntimeError("invalid source dimensions")

                 target_width, target_height = self._scaled_dimensions(width, height, max_width, max_height)
                 fps = source_reader.get("fps", {"num": 30, "den": 1})
                 pixel_ratio = source_reader.get("pixel_ratio", {"num": 1, "den": 1})
                 writer = openshot.FFmpegWriter(output_path)
                 writer.SetVideoOptions(
                     True,
                     "libx264",
                     openshot.Fraction(int(fps.get("num", 30)), int(fps.get("den", 1))),
                     target_width,
                     target_height,
                     openshot.Fraction(int(pixel_ratio.get("num", 1)), int(pixel_ratio.get("den", 1))),
                     False,
                     False,
                     28,
                 )
                 writer.PrepareStreams()

                 if source_reader.get("has_audio"):
                     channel_layout = openshot.LAYOUT_STEREO if int(source_reader.get("channels", 2) or 2) > 1 else openshot.LAYOUT_MONO
                     writer.SetAudioOptions(
                         True,
                         "aac",
                         48000,
                         2 if channel_layout == openshot.LAYOUT_STEREO else 1,
                         channel_layout,
                         128000,
                     )
                     writer.PrepareStreams()

                 writer.Open()
                 try:
                     max_frame = int(source_reader.get("video_length", 0) or 0)
                     if max_frame <= 0:
                         raise RuntimeError("invalid source frame count")
                     prewarm_frames = self._thumbnail_prewarm_frames(
                         file_id,
                         max_frame,
                         fps,
                         thumbs_per_second=self._thumbnail_prewarm_rate(),
                     )
                     prewarm_index = 0
                     for frame_number in range(1, max_frame + 1):
                         self._raise_if_canceled(file_id)
                         frame = reader.GetFrame(frame_number)
                         writer.WriteFrame(frame)
                         while prewarm_index < len(prewarm_frames) and prewarm_frames[prewarm_index] < frame_number:
                             prewarm_index += 1
                         if prewarm_index < len(prewarm_frames) and prewarm_frames[prewarm_index] == frame_number:
                             self._save_prewarmed_thumbnail(frame, file_id, frame_number, 0.0)
                             prewarm_index += 1
                         self._trim_optimize_caches(clip, reader, frame_number)
                         if frame_number == 1 or frame_number == max_frame or frame_number % max(1, min(12, max_frame // 40 or 1)) == 0:
                             self._update_progress(file_id, int((float(frame_number) / float(max_frame)) * 100.0))
                 finally:
                     writer.Close()
                     self._clear_cache_object(getattr(reader, "GetCache", lambda: None)())
                     self._clear_cache_object(getattr(clip, "GetCache", lambda: None)())
             finally:
                 clip.Close()
                 clip.ParentTimeline(None)
         except Exception:
             try:
                 if os.path.exists(output_path):
                     os.remove(output_path)
             except Exception:
                 pass
             raise

         return self._reader_json_for_path(output_path, file_id)

     def _proxy_lookup_from_payload(self, payload):
         if isinstance(payload, dict):
             files = payload.get("files")
             if isinstance(files, list):
                 return self._proxy_lookup_from_files(files)
         return {}, {}

     def _proxy_lookup_from_project(self):
         app = get_app()
         project = getattr(app, "project", None) if app else None
         files = project.get("files") if project else None
         if not isinstance(files, list):
             return {}, {}
         return self._proxy_lookup_from_files(files)

     def _proxy_lookup_from_files(self, files):
         proxy_lookup = {}
         path_lookup = {}
         for file_data in files or []:
             if not isinstance(file_data, dict):
                 continue
             file_id = str(file_data.get("id") or "")
             file_path = file_data.get("path")
             if file_id and file_path:
                 path_lookup[comparable_media_path(file_path)] = file_id
             proxy_reader = file_data.get("proxy_reader")
             if not isinstance(proxy_reader, dict) or not proxy_reader.get("path") or not file_id:
                 continue
             if not os.path.exists(str(proxy_reader.get("path"))):
                 continue
             proxy_data = copy.deepcopy(proxy_reader)
             proxy_data["id"] = file_id
             proxy_lookup[file_id] = proxy_data
         return proxy_lookup, path_lookup

     def _rewrite_payload_in_place(self, payload, proxy_lookup, path_lookup):
         changed = False
         if isinstance(payload, dict):
             for key, value in list(payload.items()):
                 if key in ("reader", "mask_reader") and isinstance(value, dict):
                     replacement = self._proxy_replacement_for_reader(value, proxy_lookup, path_lookup)
                     if replacement is not None and replacement != value:
                         payload[key] = copy.deepcopy(replacement)
                         changed = True
                 else:
                     changed = self._rewrite_payload_in_place(value, proxy_lookup, path_lookup) or changed
         elif isinstance(payload, list):
             for item in payload:
                 changed = self._rewrite_payload_in_place(item, proxy_lookup, path_lookup) or changed
         return changed

     def _proxy_replacement_for_reader(self, reader_data, proxy_lookup, path_lookup):
         reader_id = str(reader_data.get("id") or "")
         if reader_id and reader_id in proxy_lookup:
             return proxy_lookup[reader_id]

         reader_path = reader_data.get("path")
         if reader_path:
             file_id = path_lookup.get(comparable_media_path(reader_path))
             if file_id and file_id in proxy_lookup:
                 return proxy_lookup[file_id]
         return None

     def _payload_references_file(self, payload, file_id):
         file_obj = File.get(id=file_id)
         if not file_obj:
             return False
         target_path = comparable_media_path((file_obj.data or {}).get("path"))
         return self._payload_contains_file_reference(payload, str(file_id), target_path)

     def _payload_contains_file_reference(self, payload, file_id, target_path):
         if isinstance(payload, dict):
             for key, value in payload.items():
                 if key in ("reader", "mask_reader") and isinstance(value, dict):
                     if str(value.get("id") or "") == file_id:
                         return True
                     reader_path = value.get("path")
                     if reader_path and comparable_media_path(reader_path) == target_path:
                         return True
                 if self._payload_contains_file_reference(value, file_id, target_path):
                     return True
         elif isinstance(payload, list):
             for item in payload:
                 if self._payload_contains_file_reference(item, file_id, target_path):
                     return True
         return False

     def _show_status(self, text, timeout=5000):
         status_bar = getattr(self.win, "statusBar", None)
         if status_bar and hasattr(status_bar, "showMessage"):
             status_bar.showMessage(str(text), int(timeout))

     def _index_existing_optimized_files(self, folder_path):
         basename_index = {}
         stem_index = {}
         normalized_index = {}
         path_index = {}
         for root, _, files in os.walk(str(folder_path or "")):
             for name in files:
                 ext = os.path.splitext(name)[1].lower()
                 if ext not in self.MATCHABLE_VIDEO_EXTENSIONS:
                     continue
                 full_path = os.path.join(root, name)
                 basename_index.setdefault(name.lower(), []).append(full_path)
                 stem = os.path.splitext(name)[0].lower()
                 stem_index.setdefault(stem, []).append(full_path)
                 normalized_index.setdefault(self._normalized_proxy_name_key(stem), []).append(full_path)
                 path_index[comparable_media_path(full_path)] = full_path
         return {
             "basename": basename_index,
             "stem": stem_index,
             "normalized": normalized_index,
             "path": path_index,
         }

     def _match_existing_optimized_path(self, file_obj, folder_path, folder_index):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         source_path = absolute_media_path(data.get("path"))
         file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
         basename = os.path.basename(source_path or "")
         basename_lower = basename.lower()
         source_stem = os.path.splitext(basename_lower)[0]
         source_norm = comparable_media_path(source_path)

         if file_id:
             id_matches = folder_index.get("stem", {}).get(file_id.lower(), [])
             if id_matches:
                 return id_matches[0]

         if basename_lower:
             basename_matches = folder_index.get("basename", {}).get(basename_lower, [])
             if basename_matches:
                 return basename_matches[0]

         if source_stem:
             stem_matches = folder_index.get("stem", {}).get(source_stem, [])
             if stem_matches:
                 return stem_matches[0]

         source_variants = self._source_proxy_name_candidates(source_stem)
         candidate_scores = {}
         for variant in source_variants:
             for match_path in folder_index.get("stem", {}).get(variant, []):
                 candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 90)

         normalized_key = self._normalized_proxy_name_key(source_stem)
         if normalized_key:
             for match_path in folder_index.get("normalized", {}).get(normalized_key, []):
                 candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 80)

         source_prefixes = self._source_stem_file_id_prefix_candidates(source_stem, file_id)
         for candidate_stem, candidate_paths in folder_index.get("stem", {}).items():
             if source_prefixes and any(candidate_stem.startswith(prefix) for prefix in source_prefixes):
                 for match_path in candidate_paths:
                     candidate_scores[match_path] = max(candidate_scores.get(match_path, 0), 75)

         for candidate_path in folder_index.get("path", {}).values():
             if source_norm and comparable_media_path(candidate_path).endswith(source_norm):
                 candidate_scores[candidate_path] = max(candidate_scores.get(candidate_path, 0), 70)

         if candidate_scores:
             ranked_paths = sorted(
                 candidate_scores.items(),
                 key=lambda item: (
                     -item[1],
                     abs(len(os.path.splitext(os.path.basename(item[0]))[0]) - len(source_stem)),
                     os.path.basename(item[0]).lower(),
                 ),
             )
             return ranked_paths[0][0]

         expected_name = self._preferred_proxy_filename(file_id, data, folder_path, existing_names=set())
         return os.path.join(str(folder_path or ""), expected_name)

     def _missing_proxy_reader(self, file_obj, folder_path):
         data = getattr(file_obj, "data", {}) if file_obj else {}
         file_id = str(getattr(file_obj, "id", "") or data.get("id") or "")
         source_path = absolute_media_path(data.get("path"))
         basename = os.path.basename(source_path or "")
         expected_path = self._match_existing_optimized_path(file_obj, folder_path, {"basename": {}, "stem": {}, "path": {}})
         return {
             "id": file_id,
             "path": expected_path,
             "missing": True,
             "name": basename,
         }

     def _proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         os.makedirs(proxy_root, exist_ok=True)
         filename = self._preferred_proxy_filename(file_id, file_data, proxy_root)
         return os.path.join(proxy_root, filename)

     def _existing_proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         default_filename = self._preferred_proxy_filename(file_id, file_data, proxy_root, existing_names=set())
         default_path = os.path.join(proxy_root, default_filename)
         source_stem = os.path.splitext(os.path.basename(absolute_media_path((file_data or {}).get("path")) or ""))[0]
         specific_path = os.path.join(proxy_root, "{}_{}.mp4".format(self._proxy_filename_stem(source_stem), str(file_id or "")))

         if os.path.exists(specific_path):
             return specific_path
         if os.path.exists(default_path):
             return default_path
         return None

     def _reserve_proxy_output_path(self, file_id, file_data):
         proxy_root = self._proxy_root()
         os.makedirs(proxy_root, exist_ok=True)
         try:
             existing_names = set(os.listdir(proxy_root))
         except Exception:
             existing_names = set()
         with self._lock:
             for job in self._jobs.values():
                 output_path = job.get("output_path")
                 if output_path:
                     existing_names.add(os.path.basename(str(output_path)))
         filename = self._preferred_proxy_filename(file_id, file_data, proxy_root, existing_names=existing_names)
         return os.path.join(proxy_root, filename)

     def _reserved_or_computed_proxy_output_path(self, file_id, file_data):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             reserved_path = job.get("output_path") if isinstance(job, dict) else None
         return reserved_path or self._proxy_output_path(file_id, file_data)

     def _preferred_proxy_filename(self, file_id, file_data, proxy_root, existing_names=None):
         file_id = str(file_id or "")
         data = file_data or {}
         source_path = absolute_media_path(data.get("path"))
         source_stem = os.path.splitext(os.path.basename(source_path or ""))[0]
         base_stem = self._proxy_filename_stem(source_stem)
         default_name = "{}.mp4".format(base_stem)

         if existing_names is None:
             try:
                 existing_names = set(os.listdir(proxy_root))
             except Exception:
                 existing_names = set()
         existing_names_lower = {str(name).lower() for name in existing_names}

         if default_name.lower() not in existing_names_lower:
             return default_name
         return "{}_{}.mp4".format(base_stem, file_id)

     @staticmethod
     def _proxy_filename_stem(source_stem):
         source_stem = str(source_stem or "").strip()
         if not source_stem:
             return "optimized_proxy"
         return "{}_proxy".format(source_stem)


     @staticmethod
     def _source_stem_file_id_prefix_candidates(source_stem, file_id):
         source_stem = str(source_stem or "").strip().lower()
         file_id = str(file_id or "").strip().lower()
         if not source_stem or not file_id:
             return []
         separators = ("_", "-", ".")
         prefixes = []
         for separator in separators:
             prefixes.append("{}{}{}".format(source_stem, separator, file_id))
             prefixes.append("{}{}proxy{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}preview{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}optimized{}{}".format(source_stem, separator, separator, file_id))
             prefixes.append("{}{}lowres{}{}".format(source_stem, separator, separator, file_id))
         return prefixes

     @classmethod
     def _normalized_proxy_name_key(cls, stem):
         normalized = str(stem or "").strip().lower()
         if not normalized:
             return ""
         normalized = re.sub(r"[\s.-]+", "_", normalized)
         normalized = re.sub(r"_+", "_", normalized).strip("_")
         previous = None
         while normalized and normalized != previous:
             previous = normalized
             normalized = cls._PROXY_PREFIX_RE.sub("", normalized).strip("._- ")
             normalized = cls._PROXY_SUFFIX_RE.sub("", normalized).strip("._- ")
             normalized = cls._RESOLUTION_SUFFIX_RE.sub("", normalized).strip("._- ")
             if cls._CAMERA_PROXY_SUFFIX_RE.search(normalized):
                 trimmed = cls._CAMERA_PROXY_SUFFIX_RE.sub("", normalized).strip("._- ")
                 if trimmed:
                     normalized = trimmed
         return normalized

     @classmethod
     def _source_proxy_name_candidates(cls, source_stem):
         source_stem = str(source_stem or "").strip().lower()
         if not source_stem:
             return []

         candidates = {
             source_stem,
             "{}_proxy".format(source_stem),
             "{}-proxy".format(source_stem),
             "{}_optimized".format(source_stem),
             "{}-optimized".format(source_stem),
             "{}_preview".format(source_stem),
             "{}-preview".format(source_stem),
             "{}_lowres".format(source_stem),
             "{}-lowres".format(source_stem),
             "{}_p".format(source_stem),
             "{}-p".format(source_stem),
             "proxy_{}".format(source_stem),
             "proxy-{}".format(source_stem),
             "optimized_{}".format(source_stem),
             "optimized-{}".format(source_stem),
             "preview_{}".format(source_stem),
             "preview-{}".format(source_stem),
         }
         for size_label in ("360p", "540p", "720p", "1080p", "1440p", "2160p", "4k"):
             candidates.add("{}_{}".format(source_stem, size_label))
             candidates.add("{}_proxy_{}".format(source_stem, size_label))
             candidates.add("{}_{}_proxy".format(source_stem, size_label))
         return sorted(candidates)

     def _thumbnail_prewarm_frames(self, file_id, max_frame, fps, thumbs_per_second=None):
         max_frame = max(1, int(max_frame or 1))
         fps_num = float((fps or {}).get("num", 0.0) or 0.0)
         fps_den = float((fps or {}).get("den", 1.0) or 1.0)
         fps_value = (fps_num / fps_den) if fps_num > 0.0 and fps_den > 0.0 else 0.0
         if fps_value <= 0.0:
             return [1]

         if thumbs_per_second is None:
             thumbs_per_second = self._thumbnail_prewarm_rate()
         thumbs_per_second = max(1, int(thumbs_per_second))
         step = max(1, int(round(fps_value / float(thumbs_per_second))))
         frames = []
         seen = set()
         for frame_number in range(1, max_frame + 1, step):
             rounded = min(max_frame, RoundFrameToThumbnailGrid(frame_number, fps_value))
             if rounded not in seen:
                 seen.add(rounded)
                 frames.append(rounded)
         if max_frame not in seen:
             frames.append(max_frame)
         if 1 not in seen:
             frames.insert(0, 1)
         return sorted(set(frames))

     def _save_prewarmed_thumbnail(self, frame, file_id, frame_number, rotate):
         thumb_path = ThumbnailPathForFrame(file_id, frame_number)
         if os.path.exists(thumb_path):
             return
         try:
             mask_path = os.path.join(info.IMAGES_PATH, "mask.png")
             GenerateThumbnailFromFrame(frame, thumb_path, 98, 64, mask_path, "", rotate=rotate)
         except Exception:
             log.debug("Optimize Preview thumbnail prewarm failed file=%s frame=%s", file_id, frame_number, exc_info=1)

     def _configure_optimize_caches(self, clip, reader):
         self._configure_cache_object(getattr(clip, "GetCache", lambda: None)())
         self._configure_cache_object(getattr(reader, "GetCache", lambda: None)())

     def _trim_optimize_caches(self, clip, reader, frame_number):
         if int(frame_number or 0) % int(self.OPTIMIZE_CACHE_CLEAR_INTERVAL) != 0:
             return
         self._clear_cache_object(getattr(reader, "GetCache", lambda: None)())
         self._clear_cache_object(getattr(clip, "GetCache", lambda: None)())

     def _configure_cache_object(self, cache_object):
         if not cache_object:
             return
         try:
             cache_object.SetMaxBytes(int(self.OPTIMIZE_CACHE_MAX_BYTES))
         except Exception:
             log.debug("Optimize Preview cache max-bytes update failed", exc_info=1)
         self._clear_cache_object(cache_object)

     @staticmethod
     def _clear_cache_object(cache_object):
         if not cache_object:
             return
         try:
             cache_object.Clear()
         except Exception:
             log.debug("Optimize Preview cache clear failed", exc_info=1)

     def _create_optimize_timeline(self, width, height):
         width = max(2, int(width or 2))
         height = max(2, int(height or 2))
         timeline = openshot.Timeline(
             width,
             height,
             openshot.Fraction(30, 1),
             48000,
             2,
             openshot.LAYOUT_STEREO,
         )
         timeline.preview_width = width
         timeline.preview_height = height
         return timeline

     def _ensure_executor(self):
         worker_count = self._optimize_worker_count()
         current_executor = getattr(self, "_executor", None)
         current_count = getattr(current_executor, "_max_workers", None)
         if current_executor and current_count == worker_count:
             return
         with self._lock:
             if self._jobs:
                 return
             old_executor = self._executor
             self._executor = ThreadPoolExecutor(
                 max_workers=worker_count,
                 thread_name_prefix="optimized-preview",
             )
         if old_executor:
             self._shutdown_executor(old_executor)

     @staticmethod
     def _shutdown_executor(executor):
         if not executor:
             return
         try:
             executor.shutdown(wait=False, cancel_futures=True)
         except TypeError:
             executor.shutdown(wait=False)

     def _optimize_worker_count(self):
         value = self._setting_value("optimize-preview-jobs", 1)
         try:
             return max(1, min(8, int(value)))
         except (TypeError, ValueError):
             return 1

     def _thumbnail_prewarm_rate(self):
         value = self._setting_value("optimize-preview-thumbnails", 4)
         try:
             return max(1, min(12, int(value)))
         except (TypeError, ValueError):
             return 4

     def _max_optimize_bounds(self):
         raw_value = str(self._setting_value("optimize-preview-max-size", "1280x720") or "1280x720").lower()
         try:
             width_text, height_text = raw_value.split("x", 1)
             width = max(2, int(width_text))
             height = max(2, int(height_text))
             return width, height
         except (AttributeError, TypeError, ValueError):
             return 1280, 720

     @staticmethod
     def _setting_value(name, default):
         app = get_app()
         settings = app.get_settings() if app else None
         if not settings:
             return default
         try:
             value = settings.get(name)
         except Exception:
             return default
         return default if value is None else value

     def _mark_running(self, file_id):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if not job:
                 return
             if job.get("cancel_requested"):
                 job["status"] = "canceling"
             else:
                 job["status"] = "running"
             job["progress"] = max(1, int(job.get("progress", 0)))
         self._emit_job_change(file_id)

     def _update_progress(self, file_id, progress):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if not job:
                 return
             if job.get("status") not in self.ACTIVE_STATES:
                 return
             job["progress"] = max(0, min(100, int(progress)))
         self._emit_job_change(file_id)

     def _raise_if_canceled(self, file_id):
         with self._lock:
             job = self._jobs.get(str(file_id or ""))
             if job and job.get("cancel_requested"):
                 raise _ProxyJobCanceled()

     def _finalize_job(self, file_id, status):
         file_id = str(file_id or "")
         with self._lock:
             job = self._jobs.pop(file_id, None)
         if status == "completed":
             self.job_finished.emit(file_id, status)
         elif job is not None:
             self.job_finished.emit(file_id, status)
         self._emit_job_change(file_id)

     def _emit_job_change(self, file_id):
         file_id = str(file_id or "")
         if not file_id:
             return
         job = self.get_active_job_for_file(file_id)
         if job:
             self.job_updated.emit(file_id, str(job.get("status") or ""), int(job.get("progress", 0)))
         self.file_job_changed.emit(file_id)
         self.queue_changed.emit()

     @staticmethod
     def _job_snapshot(job):
         if not isinstance(job, dict):
             return None
         return {
             "id": str(job.get("id") or ""),
             "status": str(job.get("status") or ""),
             "progress": int(job.get("progress", 0)),
             "cancel_requested": bool(job.get("cancel_requested")),
         }

     @staticmethod
     def _scaled_dimensions(width, height, max_width=1280, max_height=720):
         if width <= 0 or height <= 0:
             return (1280, 720)
         scale = min(float(max_width) / float(width), float(max_height) / float(height), 1.0)
         scaled_w = max(2, int(round(width * scale)))
         scaled_h = max(2, int(round(height * scale)))
         if scaled_w % 2:
             scaled_w -= 1
         if scaled_h % 2:
             scaled_h -= 1
         return (max(2, scaled_w), max(2, scaled_h))


class _ProxyJobCanceled(RuntimeError):
     """Raised when a queued/running proxy job is canceled."""



"""
 @file
 @brief Unit tests for timeline retime keyframe scaling
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import types
import unittest
from unittest.mock import patch

try:
    import openshot
except ModuleNotFoundError:
    openshot = types.SimpleNamespace(LINEAR=1)
    sys.modules["openshot"] = openshot

classes_app = types.ModuleType("classes.app")
classes_app.get_app = lambda: None
sys.modules.setdefault("classes.app", classes_app)


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from windows.views.retime import retime_clip  # noqa: E402


def keyframe(*frames):
    return {
        "Points": [
            {"co": {"X": float(frame), "Y": float(index)}, "interpolation": openshot.LINEAR}
            for index, frame in enumerate(frames)
        ]
    }


def frame_numbers(data):
    return [point["co"]["X"] for point in data["Points"]]


class DummyClip:
    def __init__(self, data):
        self.data = data


class RetimeTests(unittest.TestCase):
    def test_retime_scales_nested_colorgrade_wheel_and_curve_keyframes(self):
        clip = DummyClip({
            "id": "clip-1",
            "position": 0.0,
            "start": 0.0,
            "end": 2.0,
            "duration": 2.0,
            "alpha": keyframe(1, 30, 60),
            "time": keyframe(1, 60),
            "effects": [{
                "class_name": "ColorGrade",
                "wheels": {
                    "enabled_keyframes": keyframe(1, 30, 60),
                    "global": {
                        "color_keyframes": {
                            "red": keyframe(1, 30, 60),
                            "green": keyframe(1, 30, 60),
                            "blue": keyframe(1, 30, 60),
                            "alpha": keyframe(1, 30, 60),
                        },
                        "amount_keyframes": keyframe(1, 30, 60),
                        "luma_keyframes": keyframe(1, 30, 60),
                    },
                },
                "curve": {
                    "enabled": keyframe(1, 30, 60),
                    "nodes": [{
                        "id": 1,
                        "x": keyframe(1, 30, 60),
                        "y": {
                            "Points": [
                                {"co": {"X": 1.0, "Y": 0.2}, "interpolation": openshot.LINEAR},
                                {"co": {"X": 30.0, "Y": 0.4}, "interpolation": openshot.LINEAR},
                                {"co": {"X": 60.0, "Y": 0.8}, "interpolation": openshot.LINEAR},
                            ]
                        },
                        "left_handle_x": keyframe(1, 30, 60),
                        "right_handle_y": keyframe(1, 30, 60),
                    }],
                },
            }],
        })

        with patch("windows.views.retime._project_fps_float", return_value=30.0):
            self.assertTrue(retime_clip(clip, 4.0, 0.0, direction=1))

        self.assertEqual(frame_numbers(clip.data["alpha"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(clip.data["time"]), [1, 121])

        effect = clip.data["effects"][0]
        wheels = effect["wheels"]
        self.assertEqual(frame_numbers(wheels["enabled_keyframes"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(wheels["global"]["color_keyframes"]["red"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(wheels["global"]["color_keyframes"]["green"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(wheels["global"]["amount_keyframes"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(wheels["global"]["luma_keyframes"]), [1.0, 60, 121])

        curve = effect["curve"]
        node = curve["nodes"][0]
        self.assertEqual(frame_numbers(curve["enabled"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(node["x"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(node["y"]), [1.0, 60, 121])
        self.assertEqual([point["co"]["Y"] for point in node["y"]["Points"]], [0.2, 0.4, 0.8])
        self.assertEqual(frame_numbers(node["left_handle_x"]), [1.0, 60, 121])
        self.assertEqual(frame_numbers(node["right_handle_y"]), [1.0, 60, 121])
        self.assertEqual(clip.data["end"], 4.0)
        self.assertEqual(clip.data["duration"], 4.0)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Reusable Film Grain preset payload helpers
"""

import copy


FILM_GRAIN_CLASS_NAME = "FilmGrain"

FILM_GRAIN_PRESET_NONE = "none"
FILM_GRAIN_PRESET_35MM_FINE = "35mm_fine"
FILM_GRAIN_PRESET_35MM_CLASSIC = "35mm_classic"
FILM_GRAIN_PRESET_35MM_GRITTY = "35mm_gritty"
FILM_GRAIN_PRESET_16MM_CLASSIC = "16mm_classic"
FILM_GRAIN_PRESET_SUPER_8 = "super_8"
FILM_GRAIN_PRESET_HIGH_ISO = "high_iso"


def _constant_property(value):
    return {
        "Points": [
            {
                "co": {"X": 1.0, "Y": float(value)},
                "handle_left": {"X": 0.5, "Y": 1.0},
                "handle_right": {"X": 0.5, "Y": 0.0},
                "handle_type": 0,
                "interpolation": 0,
            }
        ]
    }


def _set_scalar(effect_json, key, value):
    effect_json[key] = _constant_property(value)


def is_film_grain_effect(effect_json):
    if not isinstance(effect_json, dict):
        return False
    return effect_json.get("class_name") == FILM_GRAIN_CLASS_NAME


def _base_values():
    return {
        "amount": 0.25,
        "size": 0.20,
        "softness": 0.25,
        "clump": 0.20,
        "shadows": 0.80,
        "midtones": 1.00,
        "highlights": 0.55,
        "color_amount": 0.20,
        "color_variation": 0.35,
        "evolution": 0.65,
        "coherence": 0.55,
    }


def apply_film_grain_preset(effect_json, preset_name):
    payload = copy.deepcopy(effect_json or {})

    values = _base_values()
    if preset_name == FILM_GRAIN_PRESET_35MM_FINE:
        values.update({
            "amount": 0.14,
            "size": 0.12,
            "softness": 0.35,
            "clump": 0.10,
            "shadows": 0.65,
            "midtones": 0.70,
            "highlights": 0.35,
            "color_amount": 0.08,
            "color_variation": 0.20,
            "evolution": 0.45,
            "coherence": 0.75,
        })
    elif preset_name == FILM_GRAIN_PRESET_35MM_CLASSIC:
        values.update({
            "amount": 0.24,
            "size": 0.18,
            "softness": 0.22,
            "clump": 0.18,
            "shadows": 0.85,
            "midtones": 1.00,
            "highlights": 0.55,
            "color_amount": 0.16,
            "color_variation": 0.30,
            "evolution": 0.60,
            "coherence": 0.62,
        })
    elif preset_name == FILM_GRAIN_PRESET_35MM_GRITTY:
        values.update({
            "amount": 0.34,
            "size": 0.32,
            "softness": 0.28,
            "clump": 0.30,
            "shadows": 0.95,
            "midtones": 1.00,
            "highlights": 0.62,
            "color_amount": 0.20,
            "color_variation": 0.38,
            "evolution": 0.66,
            "coherence": 0.56,
        })
    elif preset_name == FILM_GRAIN_PRESET_16MM_CLASSIC:
        values.update({
            "amount": 0.42,
            "size": 0.46,
            "softness": 0.38,
            "clump": 0.44,
            "shadows": 1.00,
            "midtones": 1.00,
            "highlights": 0.70,
            "color_amount": 0.28,
            "color_variation": 0.48,
            "evolution": 0.75,
            "coherence": 0.45,
        })
    elif preset_name == FILM_GRAIN_PRESET_SUPER_8:
        values.update({
            "amount": 0.62,
            "size": 0.72,
            "softness": 0.50,
            "clump": 0.70,
            "shadows": 1.00,
            "midtones": 0.95,
            "highlights": 0.85,
            "color_amount": 0.42,
            "color_variation": 0.65,
            "evolution": 0.88,
            "coherence": 0.32,
        })
    elif preset_name == FILM_GRAIN_PRESET_HIGH_ISO:
        values.update({
            "amount": 0.52,
            "size": 0.24,
            "softness": 0.18,
            "clump": 0.24,
            "shadows": 1.00,
            "midtones": 0.95,
            "highlights": 0.42,
            "color_amount": 0.55,
            "color_variation": 0.78,
            "evolution": 0.82,
            "coherence": 0.38,
        })
    else:
        raise ValueError("Unknown film grain preset: {}".format(preset_name))

    for key, value in values.items():
        _set_scalar(payload, key, value)

    if "seed" not in payload:
        payload["seed"] = 1

    return payload



"""
 @file
 @brief Tests for recording preview and live thumbnail helpers.
"""

import os
import shutil
import sys
import tempfile
import threading
import types
import unittest
from unittest.mock import MagicMock, call, patch


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from qt_api import QByteArray, QCoreApplication, QEvent, QPoint, QRect, QSize, Qt
from qt_api import QApplication

from tests.qt_test_app import ensure_app_state as ensure_qt_app_state, get_or_create_app

QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)


class DummySettings:
    def __init__(self):
        self.values = {
            "default-profile": "HD 720p 30 fps",
            "default-samplerate": 48000,
            "default-channels": 2,
            "default-image-length": 10.0,
        }

    def get(self, key):
        return self.values.get(key)


class DummyApp(QApplication):
    def __init__(self):
        super().__init__([])
        self.settings = DummySettings()

    def get_settings(self):
        return self.settings

    def _tr(self, text):
        return text


def ensure_app_state(app):
    return ensure_qt_app_state(app, DummySettings)


class RecordingPreviewTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        app, cls._owns_app = get_or_create_app(DummyApp)
        cls.app = ensure_app_state(app)
        import windows.audio_recording as audio_recording_module
        import windows.recording_widgets as recording_widgets_module
        import windows.views.timeline as timeline_module
        cls.audio_recording_module = audio_recording_module
        cls.recording_widgets_module = recording_widgets_module
        cls.timeline_module = timeline_module

    @classmethod
    def tearDownClass(cls):
        if getattr(cls, "_owns_app", False) and cls.app:
            cls.app.quit()

    def test_recording_preview_file_id_sanitizes_source(self):
        helper = self.audio_recording_module

        self.assertEqual(
            helper.recording_preview_file_id("session-1", "screen capture"),
            "recording-preview-session-1-screen-capture",
        )
        self.assertEqual(
            helper.recording_preview_file_id("session-1", ""),
            "recording-preview-session-1-source",
        )

    def test_recording_styles_are_owned_by_application_themes(self):
        from themes.cosmic.theme import CosmicTheme
        from themes.humanity.theme import HumanityDarkTheme, Retro

        cosmic = CosmicTheme(self.app).style_sheet
        humanity = HumanityDarkTheme(self.app).style_sheet
        retro = Retro(self.app).style_sheet

        self.assertIn("QFrame#recordingCard", cosmic)
        self.assertIn("background-color: #303030", humanity)
        self.assertIn("background-color: #e5e7ea", retro)

        card = self.recording_widgets_module.RecordingSourceCard("Mic", "Voice", "M")
        section = self.recording_widgets_module.RecordingSection("Mic", "M")
        try:
            self.assertEqual(card.styleSheet(), "")
            self.assertEqual(section.styleSheet(), "")
        finally:
            card.deleteLater()
            section.deleteLater()

    def test_recording_dock_starts_with_microphone_unselected(self):
        helper = self.audio_recording_module
        content_class = helper.AudioRecordingDockContent
        with patch.object(content_class, "_sync_source_availability"), \
                patch.object(content_class, "_webcam_layout_changed"), \
                patch.object(content_class, "_sync_backend_state"):
            dock = content_class(types.SimpleNamespace())
        try:
            self.assertFalse(dock.mic_card.isChecked())
            self.assertFalse(dock.mic_section.property("active"))
            self.assertFalse(dock.mic_section.advanced_button.isEnabled())
        finally:
            dock.deleteLater()

    def test_recording_button_requires_selected_source_while_idle(self):
        helper = self.audio_recording_module

        class FakeButton:
            def setEnabled(self, enabled):
                self.enabled = enabled

            def setText(self, text):
                self.text = text

            def setStyleSheet(self, stylesheet):
                self.stylesheet = stylesheet

            def setToolTip(self, tooltip):
                self.tooltip = tooltip

        selected = {"mic": False}
        dock = types.SimpleNamespace(
            mic_card=types.SimpleNamespace(isChecked=lambda: selected["mic"]),
            screen_card=types.SimpleNamespace(isChecked=lambda: False),
            camera_card=types.SimpleNamespace(isChecked=lambda: False),
            record_button=FakeButton(),
        )
        dock._has_selected_recording_source = lambda: (
            helper.AudioRecordingDockContent._has_selected_recording_source(dock)
        )

        helper.AudioRecordingDockContent._set_record_button_idle(dock)
        self.assertFalse(dock.record_button.enabled)
        self.assertIn("Select at least one", dock.record_button.tooltip)
        self.assertIn("QPushButton:disabled", dock.record_button.stylesheet)

        selected["mic"] = True
        helper.AudioRecordingDockContent._set_record_button_idle(dock)
        self.assertTrue(dock.record_button.enabled)
        self.assertEqual(dock.record_button.tooltip, "")

    def test_recording_dock_idle_activation_skips_device_discovery(self):
        helper = self.audio_recording_module
        dock = types.SimpleNamespace(
            _activation_pending=True,
            mic_card=types.SimpleNamespace(isChecked=lambda: False),
            camera_card=types.SimpleNamespace(isChecked=lambda: False),
            _dock_visible=lambda: True,
            refresh_tracks=MagicMock(),
            _sync_backend_state=MagicMock(),
            _refresh_source_devices=MagicMock(),
            _ensure_monitoring=MagicMock(),
            _restart_webcam_preview=MagicMock(),
        )

        helper.AudioRecordingDockContent._activate_after_show(dock)

        self.assertFalse(dock._activation_pending)
        dock.refresh_tracks.assert_called_once_with()
        dock._refresh_source_devices.assert_called_once_with(
            microphone=False,
            camera=False,
        )
        dock._ensure_monitoring.assert_called_once_with()
        dock._restart_webcam_preview.assert_called_once_with()

    def test_recording_source_discovery_runs_only_for_requested_source(self):
        helper = self.audio_recording_module
        dock = types.SimpleNamespace(
            _set_wait_cursor=MagicMock(),
            refresh_devices=MagicMock(),
            _sync_channel_options=MagicMock(),
            refresh_cameras=MagicMock(),
        )

        helper.AudioRecordingDockContent._refresh_source_devices(
            dock, microphone=True
        )

        dock.refresh_devices.assert_called_once_with()
        dock._sync_channel_options.assert_called_once_with()
        dock.refresh_cameras.assert_not_called()
        self.assertEqual(
            dock._set_wait_cursor.call_args_list,
            [call(True), call(False)],
        )

    def test_webcam_source_stays_selectable_until_lazy_discovery_runs(self):
        helper = self.audio_recording_module
        dock = types.SimpleNamespace(
            mic_card=types.SimpleNamespace(setAvailable=MagicMock()),
            screen_card=types.SimpleNamespace(setAvailable=MagicMock()),
            camera_card=types.SimpleNamespace(setAvailable=MagicMock()),
            _screen_backend_available=lambda: False,
            _camera_backend_available=lambda: True,
            _camera_device_available=lambda: False,
            _sync_screen_backend_ui=MagicMock(),
            _camera_devices_refreshed=False,
        )

        helper.AudioRecordingDockContent._sync_source_availability(dock)

        dock.camera_card.setAvailable.assert_called_once_with(True, "")

        dock.camera_card.setAvailable.reset_mock()
        dock._camera_devices_refreshed = True
        helper.AudioRecordingDockContent._sync_source_availability(dock)

        available, tooltip = dock.camera_card.setAvailable.call_args.args
        self.assertFalse(available)
        self.assertIn("No webcam", tooltip)

    def test_webcam_preview_stop_closes_reader_before_joining_worker(self):
        helper = self.audio_recording_module
        released = helper.threading.Event()

        class BlockingReader:
            def __init__(self):
                self.closed = False

            def Close(self):
                self.closed = True
                released.set()

        reader = BlockingReader()
        job = helper.WebcamPreviewJob("", 640, 480)
        job._reader = reader
        job._thread = helper.threading.Thread(target=released.wait, daemon=True)
        job._thread.start()

        job.stop()

        self.assertTrue(reader.closed)
        self.assertFalse(job._thread.is_alive())

    def test_xdotool_window_picker_rejects_invalid_window_id(self):
        helper = self.recording_widgets_module
        selected = types.SimpleNamespace(stdout="123;touch /tmp/not-safe\n")

        with patch.object(helper.shutil, "which", return_value="/usr/bin/xdotool"), \
                patch.object(helper.subprocess, "run", return_value=selected) as run:
            self.assertIsNone(helper.pick_x11_window_with_xdotool())

        run.assert_called_once_with(
            ["xdotool", "selectwindow"],
            check=True,
            text=True,
            capture_output=True,
            timeout=30,
        )

    def test_screen_capture_backend_uses_libopenshot_default_backend(self):
        helper = self.audio_recording_module

        class FakeReader:
            @staticmethod
            def DefaultBackend():
                return 2

            @staticmethod
            def IsBackendSupported(backend):
                return backend == 2

        fake_openshot = types.SimpleNamespace(
            ScreenCaptureReader=FakeReader,
            ScreenCaptureSettings=object,
            SCREEN_CAPTURE_AUTO=0,
            SCREEN_CAPTURE_X11=1,
            SCREEN_CAPTURE_WAYLAND=2,
        )

        with patch.object(helper.sys, "platform", "linux"), patch.object(helper, "openshot", fake_openshot):
            self.assertEqual(helper.screen_capture_backend(), 2)
            self.assertTrue(helper.screen_capture_backend_is_wayland())
            self.assertTrue(helper.screen_capture_backend_supported())

    def test_screen_capture_auto_backend_does_not_enable_screen_source(self):
        helper = self.audio_recording_module

        class FakeReader:
            @staticmethod
            def DefaultBackend():
                return 0

            @staticmethod
            def IsBackendSupported(backend):
                return backend == 0

        fake_openshot = types.SimpleNamespace(
            ScreenCaptureReader=FakeReader,
            ScreenCaptureSettings=object,
            SCREEN_CAPTURE_AUTO=0,
            SCREEN_CAPTURE_X11=1,
            SCREEN_CAPTURE_WAYLAND=2,
        )

        with patch.object(helper.sys, "platform", "linux"), patch.object(helper, "openshot", fake_openshot):
            self.assertEqual(helper.screen_capture_backend(), 0)
            self.assertFalse(helper.screen_capture_backend_supported())

    def test_screen_capture_backend_requires_explicit_wayland_support(self):
        helper = self.audio_recording_module

        fake_openshot = types.SimpleNamespace(
            ScreenCaptureReader=object,
            ScreenCaptureSettings=object,
            SCREEN_CAPTURE_AUTO=0,
            SCREEN_CAPTURE_X11=1,
            SCREEN_CAPTURE_WAYLAND=2,
        )

        with patch.object(helper.sys, "platform", "linux"), patch.object(helper, "openshot", fake_openshot):
            with patch.dict(helper.os.environ, {"XDG_SESSION_TYPE": "wayland"}, clear=False):
                self.assertEqual(helper.screen_capture_backend(), 2)
                self.assertFalse(helper.screen_capture_backend_supported())

    def test_windows_screen_capture_backend_uses_gdi_when_supported(self):
        helper = self.audio_recording_module

        class FakeReader:
            @staticmethod
            def DefaultBackend():
                return 3

            @staticmethod
            def IsBackendSupported(backend):
                return backend == 3

        fake_openshot = types.SimpleNamespace(
            ScreenCaptureReader=FakeReader,
            ScreenCaptureSettings=object,
            SCREEN_CAPTURE_AUTO=0,
            SCREEN_CAPTURE_X11=1,
            SCREEN_CAPTURE_WAYLAND=2,
            SCREEN_CAPTURE_WINDOWS_GDI=3,
        )

        with patch.object(helper.sys, "platform", "win32"), patch.object(helper, "openshot", fake_openshot):
            self.assertEqual(helper.screen_capture_backend(), 3)
            self.assertTrue(helper.screen_capture_backend_is_windows())
            self.assertTrue(helper.screen_capture_backend_supported())

    def test_live_thumbnail_cache_saves_only_grid_frames_once(self):
        helper = self.audio_recording_module
        temp_dir = tempfile.mkdtemp()
        calls = []

        def fake_save(self, frame, path):
            calls.append((frame, path, self.thumb_width, self.thumb_height))
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as output:
                output.write(b"png")

        try:
            with patch.object(helper.info, "THUMBNAIL_PATH", temp_dir), patch.object(
                helper.LiveRecordingThumbnailCache,
                "_save_thumbnail_from_frame",
                fake_save,
            ):
                cache = helper.LiveRecordingThumbnailCache(
                    "temp-file",
                    8.0,
                    thumb_size=QSize(64, 36),
                )

                self.assertTrue(cache.save_frame("frame-1", 1).endswith(os.path.join("temp-file", "1.png")))
                self.assertEqual(cache.save_frame("frame-2", 2), "")
                self.assertTrue(cache.save_frame("frame-3", 3).endswith(os.path.join("temp-file", "3.png")))
                self.assertTrue(cache.save_frame("frame-3b", 3).endswith(os.path.join("temp-file", "3.png")))

            self.assertEqual([call[0] for call in calls], ["frame-1", "frame-3"])
            self.assertEqual(calls[0][2:4], (64, 36))
        finally:
            shutil.rmtree(temp_dir, ignore_errors=True)

    def test_live_thumbnail_cache_copies_temp_folder_to_final_file_id(self):
        helper = self.audio_recording_module
        temp_dir = tempfile.mkdtemp()
        try:
            with patch.object(helper.info, "THUMBNAIL_PATH", temp_dir):
                source_dir = os.path.join(temp_dir, "temp-file")
                os.makedirs(source_dir)
                with open(os.path.join(source_dir, "1.png"), "wb") as output:
                    output.write(b"one")
                with open(os.path.join(source_dir, "note.txt"), "w") as output:
                    output.write("skip")

                copied = helper.LiveRecordingThumbnailCache("temp-file", 30.0).copy_to_file_id("final-file")

            self.assertEqual(copied, 1)
            with open(os.path.join(temp_dir, "final-file", "1.png"), "rb") as copied_file:
                self.assertEqual(copied_file.read(), b"one")
            self.assertFalse(os.path.exists(os.path.join(temp_dir, "final-file", "note.txt")))
        finally:
            shutil.rmtree(temp_dir, ignore_errors=True)

    def test_live_video_recording_preserves_elapsed_frame_numbers(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def __init__(self):
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeReader:
            def GetFrame(self, _frame_number):
                return FakeFrame()

        class FakeWriter:
            def __init__(self, job):
                self.job = job
                self.frame_numbers = []

            def WriteFrame(self, frame):
                self.frame_numbers.append(frame.number)
                if frame.number >= 31:
                    self.job._stop.set()

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(FakeReader(), "screen.mp4", 640, 480, fps)
        writer = FakeWriter(job)
        job._writer = writer
        job.begin(100.0)

        with patch.object(helper.time, "monotonic", side_effect=[100.0, 100.5, 101.0]):
            job._run()

        self.assertEqual(writer.frame_numbers, [1, 16, 31])
        self.assertEqual(job.frames, 31)

    def test_live_video_writer_enables_audio_for_screen_reader_audio(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def GetWidth(self):
                return 640

            def GetHeight(self):
                return 360

        class FakeReader:
            info = types.SimpleNamespace(
                width=640,
                height=360,
                fps=types.SimpleNamespace(num=30, den=1),
                has_audio=True,
                sample_rate=48000,
                channels=2,
            )

            def Open(self):
                pass

            def GetFrame(self, _number):
                return FakeFrame()

        class FakeWriter:
            instance = None

            def __init__(self, path):
                self.path = path
                self.audio_options = None
                FakeWriter.instance = self

            def SetVideoOptions(self, *args):
                pass

            def SetAudioOptions(self, *args):
                self.audio_options = args

            def Open(self):
                pass

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(FakeReader(), "screen.mp4", 640, 360, fps)
        with patch.object(helper.openshot, "FFmpegWriter", FakeWriter):
            job._open()

        self.assertEqual(
            FakeWriter.instance.audio_options,
            (True, "aac", 48000, 2, helper.openshot.LAYOUT_STEREO, 192000),
        )

    def test_live_video_recording_drops_frames_that_arrive_before_next_fps_slot(self):
        helper = self.audio_recording_module
        frame_times = [100.0, 100.010, 100.020, 100.0334, 100.050, 100.0667, 100.100]

        class FakeFrame:
            def __init__(self):
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeReader:
            def __init__(self):
                self.job = None
                self.frames_read = 0
                self.audio_frames = []
                self.audio_reset = False

            def GetFrame(self, _frame_number):
                self.frames_read += 1
                if self.job and self.frames_read >= len(frame_times):
                    self.job._stop.set()
                return FakeFrame()

            def ResetSystemAudio(self):
                self.audio_reset = True

            def AddSystemAudio(self, _frame, frame_number):
                self.audio_frames.append(frame_number)

        class FakeWriter:
            def __init__(self):
                self.frame_numbers = []

            def WriteFrame(self, frame):
                self.frame_numbers.append(frame.number)

        fps = types.SimpleNamespace(num=30, den=1)
        reader = FakeReader()
        job = helper.LiveVideoRecordingJob(reader, "screen.mp4", 640, 480, fps)
        reader.job = job
        writer = FakeWriter()
        job._writer = writer
        job.begin(100.0)

        with patch.object(helper.time, "monotonic", side_effect=frame_times):
            job._run()

        self.assertEqual(writer.frame_numbers, [1, 2, 3, 4])
        self.assertTrue(reader.audio_reset)
        self.assertEqual(reader.audio_frames, [1, 2, 3, 4])

    def test_live_video_recording_discards_prestart_initial_frame(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def __init__(self, label):
                self.label = label
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeReader:
            def GetFrame(self, _frame_number):
                return FakeFrame("fresh")

        class FakeWriter:
            def __init__(self, job):
                self.job = job
                self.frames = []

            def WriteFrame(self, frame):
                self.frames.append((frame.label, frame.number))
                self.job._stop.set()

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(FakeReader(), "webcam.mp4", 640, 480, fps)
        writer = FakeWriter(job)
        job._writer = writer
        job._initial_frame = FakeFrame("stale")
        job._initial_frame_time = 99.0
        job.begin(100.0)

        with patch.object(helper.time, "monotonic", return_value=100.0):
            job._run()

        self.assertEqual(writer.frames, [("fresh", 1)])

    def test_live_video_recording_drops_queued_source_frames_before_start(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def __init__(self, label, capture_timestamp):
                self.label = label
                self.capture_timestamp = capture_timestamp
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeReader:
            def __init__(self):
                self.frames = [
                    FakeFrame("queued-1", 10.033),
                    FakeFrame("queued-2", 10.066),
                    FakeFrame("live-1", 10.300),
                    FakeFrame("live-2", 10.333),
                ]

            def GetFrame(self, _frame_number):
                return self.frames.pop(0)

        class FakeWriter:
            def __init__(self, job):
                self.job = job
                self.frames = []

            def WriteFrame(self, frame):
                self.frames.append((frame.label, frame.number))
                if len(self.frames) >= 2:
                    self.job._stop.set()

        fps = types.SimpleNamespace(num=30, den=1)
        reader = FakeReader()
        job = helper.LiveVideoRecordingJob(reader, "webcam.mp4", 640, 480, fps)
        writer = FakeWriter(job)
        job._writer = writer
        job._initial_frame = FakeFrame("stale", 10.0)
        job._initial_frame_time = 100.0
        job._initial_capture_timestamp = 10.0
        job.begin(100.3)

        with patch.object(helper.time, "monotonic", return_value=100.3):
            job._run()

        self.assertEqual(writer.frames, [("live-1", 1), ("live-2", 2)])

    def test_live_video_async_closes_reader_on_worker_thread(self):
        helper = self.audio_recording_module
        calls = []

        class FakeReader:
            def Close(self):
                calls.append(("close", threading.get_ident()))

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(FakeReader(), "webcam.mp4", 640, 480, fps)
        job._open = lambda: calls.append(("open", threading.get_ident()))
        job._run = lambda: calls.append(("run", threading.get_ident()))

        caller_thread = threading.get_ident()
        job.start_async()
        job._thread.join(timeout=2.0)
        job.finish_stop()

        self.assertFalse(job._thread.is_alive())
        self.assertEqual([name for name, _thread in calls], ["open", "run", "close"])
        self.assertEqual(calls[0][1], calls[2][1])
        self.assertNotEqual(caller_thread, calls[2][1])

    def test_live_video_stop_force_closes_reader_to_unblock_and_finalizes_writer(self):
        helper = self.audio_recording_module

        class FakeReader:
            def __init__(self):
                self.closed = False

            def Close(self):
                self.closed = True

        class FakeWriter:
            def __init__(self):
                self.closed = False

            def Close(self):
                self.closed = True

        class FakeThread:
            def __init__(self):
                self.joins = 0

            def join(self, timeout=None):
                self.joins += 1

            def is_alive(self):
                return self.joins < 2

        reader = FakeReader()
        writer = FakeWriter()
        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(reader, "screen.mp4", 640, 480, fps)
        job._thread = FakeThread()
        job._writer = writer

        job.stop()

        self.assertTrue(reader.closed)
        self.assertTrue(writer.closed)
        self.assertIsNone(job._writer)

    def test_stop_video_jobs_requests_shared_stop_before_finalize(self):
        helper = self.audio_recording_module
        calls = []

        class FakeJob:
            path = "screen.mp4"
            error = None

            def __init__(self, name):
                self.name = name

            def request_stop(self, stop_time):
                calls.append(("request", self.name, stop_time))

            def finish_stop(self):
                calls.append(("finish", self.name, None))

        dock = types.SimpleNamespace(_video_jobs=[FakeJob("screen"), FakeJob("webcam")])

        helper.AudioRecordingDockContent._stop_video_jobs(dock, stop_time=123.45)

        self.assertEqual(calls[:2], [("request", "screen", 123.45), ("request", "webcam", 123.45)])
        self.assertEqual(calls[2:], [("finish", "screen", None), ("finish", "webcam", None)])
        self.assertEqual(dock._video_jobs, [])

    def test_live_video_write_uses_write_frame_at_when_available(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def __init__(self):
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeWriter:
            def __init__(self):
                self.frame_numbers = []

            def WriteFrameAt(self, frame, frame_number):
                self.frame_numbers.append((frame.number, frame_number))

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(object(), "screen.mp4", 640, 480, fps)
        writer = FakeWriter()
        job._writer = writer
        frame = FakeFrame()

        job._write_numbered_frame(frame, 42)

        self.assertEqual(writer.frame_numbers, [(42, 42)])

    def test_live_video_stop_does_not_write_final_gap_frames(self):
        helper = self.audio_recording_module

        class FakeFrame:
            def __init__(self):
                self.number = 0

            def SetFrameNumber(self, frame_number):
                self.number = frame_number

        class FakeWriter:
            def __init__(self):
                self.frame_numbers = []
                self.closed = False

            def WriteFrame(self, frame):
                self.frame_numbers.append(frame.number)

            def Close(self):
                self.closed = True

        class FakeReader:
            def Close(self):
                pass

        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(FakeReader(), "screen.mp4", 640, 480, fps)
        writer = FakeWriter()
        job._writer = writer
        job._start_time = 100.0
        job._last_output_frame_number = 31
        job._last_frame = FakeFrame()
        job._last_frame.SetFrameNumber(31)

        with patch.object(helper.time, "monotonic", return_value=110.0):
            job.stop()

        self.assertIsNone(job._writer)
        self.assertTrue(writer.closed)
        self.assertEqual(writer.frame_numbers, [])
        self.assertEqual(job.frames, 0)

    def test_live_video_wait_until_opened_times_out_and_closes_reader(self):
        helper = self.audio_recording_module

        class FakeReader:
            def __init__(self):
                self.closed = False

            def Close(self):
                self.closed = True

        reader = FakeReader()
        fps = types.SimpleNamespace(num=30, den=1)
        job = helper.LiveVideoRecordingJob(reader, "screen.mp4", 640, 480, fps)

        with self.assertRaises(RuntimeError):
            job.wait_until_opened(timeout=0.01)

        self.assertTrue(reader.closed)
        self.assertTrue(job._stop.is_set())

    def test_recording_duration_keeps_existing_media_duration(self):
        helper = self.audio_recording_module

        class FakeFile:
            def __init__(self):
                self.data = {
                    "duration": 1.0,
                    "end": 1.0,
                    "video_length": 30,
                    "has_video": True,
                    "has_audio": False,
                    "media_type": "video",
                }
                self.saved = False

            def save(self):
                self.saved = True

        fake_file = FakeFile()
        fake_app = types.SimpleNamespace(
            project=types.SimpleNamespace(get=lambda key: {"num": 30, "den": 1} if key == "fps" else None)
        )
        dock = types.SimpleNamespace(_recorded_duration=2.49)

        with patch.object(helper.File, "get", return_value=fake_file), patch.object(helper, "get_app", return_value=fake_app):
            helper.AudioRecordingDockContent._apply_recording_duration(dock, "mic.wav")

        self.assertFalse(fake_file.saved)
        self.assertEqual(fake_file.data["duration"], 1.0)
        self.assertEqual(fake_file.data["end"], 1.0)
        self.assertEqual(fake_file.data["video_length"], 30)

    def test_webcam_layout_defaults_follow_screen_selection(self):
        helper = self.audio_recording_module

        class FakeCard:
            def __init__(self, checked):
                self.checked = checked

            def isChecked(self):
                return self.checked

        class FakeCombo:
            def __init__(self, values, current):
                self.values = list(values)
                self.index = self.values.index(current)

            def findData(self, value):
                try:
                    return self.values.index(value)
                except ValueError:
                    return -1

            def currentIndex(self):
                return self.index

            def setCurrentIndex(self, index):
                self.index = index

            def currentData(self):
                return self.values[self.index]

        dock = types.SimpleNamespace(
            camera_card=FakeCard(True),
            screen_card=FakeCard(False),
            webcam_layout_combo=FakeCombo(["bottom-right", "full"], "bottom-right"),
            webcam_layout_size_combo=FakeCombo([0.2, 0.3, 0.4], 0.4),
            webcam_corner_radius_combo=FakeCombo([0.0, 0.15, 0.5], 0.15),
            _webcam_layout_default_state=None,
        )
        dock._set_combo_data = lambda combo, value: helper.AudioRecordingDockContent._set_combo_data(dock, combo, value)

        helper.AudioRecordingDockContent._sync_webcam_layout_defaults(dock)

        self.assertEqual(dock.webcam_layout_combo.currentData(), "full")
        self.assertEqual(dock.webcam_corner_radius_combo.currentData(), 0.0)
        self.assertEqual(helper.AudioRecordingDockContent._webcam_layout(dock), "full")
        self.assertEqual(helper.AudioRecordingDockContent._webcam_corner_radius(dock), 0.0)

        dock.screen_card.checked = True
        helper.AudioRecordingDockContent._sync_webcam_layout_defaults(dock)

        self.assertEqual(dock.webcam_layout_combo.currentData(), "bottom-right")
        self.assertEqual(dock.webcam_layout_size_combo.currentData(), 0.3)
        self.assertEqual(dock.webcam_corner_radius_combo.currentData(), 0.15)

        dock.webcam_layout_combo.setCurrentIndex(dock.webcam_layout_combo.findData("full"))
        helper.AudioRecordingDockContent._sync_webcam_layout_defaults(dock)

        self.assertEqual(dock.webcam_layout_combo.currentData(), "full")

    def test_screen_recording_forces_preview_off_then_restores_it(self):
        helper = self.audio_recording_module

        class FakeCard:
            checked = False

            def isChecked(self):
                return self.checked

        class FakeCombo:
            values = ["none", "full", "half", "quarter"]

            def __init__(self):
                self.index = self.values.index("half")
                self.enabled = True

            def currentData(self):
                return self.values[self.index]

            def currentIndex(self):
                return self.index

            def findData(self, value):
                return self.values.index(value)

            def setCurrentIndex(self, index):
                self.index = index

            def setEnabled(self, enabled):
                self.enabled = enabled

        card = FakeCard()
        combo = FakeCombo()
        label = types.SimpleNamespace(setEnabled=lambda enabled: setattr(label, "enabled", enabled))
        dock = types.SimpleNamespace(
            screen_card=card,
            preview_combo=combo,
            preview_label=label,
            _preview_before_screen="full",
            _preview_forced_off=False,
        )
        dock._set_combo_data = lambda target, value: helper.AudioRecordingDockContent._set_combo_data(dock, target, value)

        card.checked = True
        helper.AudioRecordingDockContent._sync_preview_control(dock)
        self.assertEqual(combo.currentData(), "none")
        self.assertFalse(combo.enabled)
        self.assertFalse(label.enabled)

        card.checked = False
        helper.AudioRecordingDockContent._sync_preview_control(dock)
        self.assertEqual(combo.currentData(), "half")
        self.assertTrue(combo.enabled)
        self.assertTrue(label.enabled)

    def test_recording_only_stops_playback_it_started(self):
        helper = self.audio_recording_module
        pause_calls = []
        dock = types.SimpleNamespace(
            _timeline_playback_started=True,
            _playback_active=lambda: True,
            window=types.SimpleNamespace(
                PauseSignal=types.SimpleNamespace(emit=lambda: pause_calls.append(True)),
            ),
        )

        helper.AudioRecordingDockContent._stop_timeline_playback(dock)
        helper.AudioRecordingDockContent._stop_timeline_playback(dock)

        self.assertEqual(pause_calls, [True])
        self.assertFalse(dock._timeline_playback_started)

    def test_recording_start_context_is_consumed_once(self):
        helper = self.audio_recording_module
        playhead = {"position": 20.0}
        dock = types.SimpleNamespace(
            _context_start=5.0,
            _recording_timeline_position=0.0,
            _current_playhead_seconds=lambda: playhead["position"],
        )

        first = helper.AudioRecordingDockContent._capture_recording_timeline_position(dock)
        self.assertEqual(first, 5.0)
        self.assertIsNone(dock._context_start)

        playhead["position"] = 30.0
        second = helper.AudioRecordingDockContent._capture_recording_timeline_position(dock)
        self.assertEqual(second, 30.0)
        self.assertEqual(dock._recording_timeline_position, 30.0)

    def test_restore_hidden_openshot_maximized_window_keeps_maximized_state(self):
        helper = self.audio_recording_module

        class FakeWindow:
            def __init__(self):
                self.calls = []

            def restoreGeometry(self, geometry):
                self.calls.append(("restoreGeometry", bytes(geometry)))

            def showFullScreen(self):
                self.calls.append("showFullScreen")

            def showMaximized(self):
                self.calls.append("showMaximized")

            def showNormal(self):
                self.calls.append("showNormal")

            def raise_(self):
                self.calls.append("raise")

            def activateWindow(self):
                self.calls.append("activateWindow")

        window = FakeWindow()
        dock = types.SimpleNamespace(window=window)
        state = {
            "window_state": Qt.WindowMaximized,
            "geometry": QByteArray(b"saved-geometry"),
        }

        helper.AudioRecordingDockContent._restore_openshot_window(dock, state)

        self.assertEqual(window.calls[:3], ["showMaximized", "raise", "activateWindow"])
        self.assertNotIn(("restoreGeometry", b"saved-geometry"), window.calls)

    def test_restore_hidden_openshot_normal_window_restores_geometry(self):
        helper = self.audio_recording_module

        class FakeWindow:
            def __init__(self):
                self.calls = []

            def restoreGeometry(self, geometry):
                self.calls.append(("restoreGeometry", bytes(geometry)))

            def showFullScreen(self):
                self.calls.append("showFullScreen")

            def showMaximized(self):
                self.calls.append("showMaximized")

            def showNormal(self):
                self.calls.append("showNormal")

            def raise_(self):
                self.calls.append("raise")

            def activateWindow(self):
                self.calls.append("activateWindow")

        window = FakeWindow()
        dock = types.SimpleNamespace(window=window)
        state = {
            "window_state": Qt.WindowNoState,
            "geometry": QByteArray(b"saved-geometry"),
        }

        helper.AudioRecordingDockContent._restore_openshot_window(dock, state)

        self.assertEqual(
            window.calls[:4],
            [("restoreGeometry", b"saved-geometry"), "showNormal", "raise", "activateWindow"],
        )

    def test_webcam_corner_layout_uses_native_margin_and_corner_radius(self):
        helper = self.audio_recording_module

        class FakeCard:
            def __init__(self, checked):
                self.checked = checked

            def isChecked(self):
                return self.checked

        class FakeCombo:
            def __init__(self, value):
                self.value = value

            def currentData(self):
                return self.value

        dock = types.SimpleNamespace(
            screen_card=FakeCard(True),
            webcam_layout_combo=FakeCombo("bottom-right"),
            webcam_layout_size_combo=FakeCombo(0.3),
            webcam_corner_radius_combo=FakeCombo(0.15),
            _keyframe_point=lambda value: {"co": {"X": 1.0, "Y": float(value)}},
        )
        dock._webcam_layout = lambda: helper.AudioRecordingDockContent._webcam_layout(dock)
        dock._webcam_layout_scale = lambda: helper.AudioRecordingDockContent._webcam_layout_scale(dock)
        dock._webcam_corner_radius = lambda: helper.AudioRecordingDockContent._webcam_corner_radius(dock)
        clip_data = {}

        helper.AudioRecordingDockContent._apply_webcam_clip_layout(dock, clip_data)

        self.assertEqual(clip_data["scale_x"]["Points"][0]["co"]["Y"], 0.3)
        self.assertEqual(clip_data["scale_y"]["Points"][0]["co"]["Y"], 0.3)
        self.assertEqual(clip_data["scale"], helper.openshot.SCALE_FIT)
        self.assertEqual(clip_data["margin"]["Points"][0]["co"]["Y"], 0.03)
        self.assertEqual(clip_data["corner_radius"]["Points"][0]["co"]["Y"], 0.15)
        self.assertNotIn("effects", clip_data)

    def test_webcam_full_layout_uses_best_fit_not_stretch(self):
        helper = self.audio_recording_module

        class FakeCard:
            def __init__(self, checked):
                self.checked = checked

            def isChecked(self):
                return self.checked

        class FakeCombo:
            def __init__(self, value):
                self.value = value

            def currentData(self):
                return self.value

        dock = types.SimpleNamespace(
            screen_card=FakeCard(False),
            webcam_layout_combo=FakeCombo("full"),
            webcam_layout_size_combo=FakeCombo(0.3),
            webcam_corner_radius_combo=FakeCombo(0.15),
            _keyframe_point=lambda value: {"co": {"X": 1.0, "Y": float(value)}},
        )
        dock._webcam_layout = lambda: helper.AudioRecordingDockContent._webcam_layout(dock)
        dock._webcam_layout_scale = lambda: helper.AudioRecordingDockContent._webcam_layout_scale(dock)
        dock._webcam_corner_radius = lambda: helper.AudioRecordingDockContent._webcam_corner_radius(dock)
        clip_data = {"scale": helper.openshot.SCALE_STRETCH}

        helper.AudioRecordingDockContent._apply_webcam_clip_layout(dock, clip_data)

        self.assertEqual(clip_data["gravity"], helper.openshot.GRAVITY_CENTER)
        self.assertEqual(clip_data["scale"], helper.openshot.SCALE_FIT)
        self.assertEqual(clip_data["scale_x"]["Points"][0]["co"]["Y"], 1.0)
        self.assertEqual(clip_data["scale_y"]["Points"][0]["co"]["Y"], 1.0)
        self.assertEqual(clip_data["margin"]["Points"][0]["co"]["Y"], 0.0)
        self.assertEqual(clip_data["corner_radius"]["Points"][0]["co"]["Y"], 0.0)

    def test_recording_clip_defaults_use_best_fit(self):
        helper = self.audio_recording_module
        dock = types.SimpleNamespace()
        clip_data = {"scale": helper.openshot.SCALE_STRETCH}

        helper.AudioRecordingDockContent._apply_recording_clip_defaults(dock, clip_data)

        self.assertEqual(clip_data["scale"], helper.openshot.SCALE_FIT)

    def test_recording_track_stack_creates_lower_tracks_when_needed(self):
        helper = self.audio_recording_module

        class FakeCombo:
            def currentData(self):
                return 2000000

        created = []
        fake_window = types.SimpleNamespace(
            create_track_below=lambda layer: created.append(max(1, int(layer / 2))) or created[-1]
        )
        fake_app = types.SimpleNamespace(window=fake_window)
        dock = types.SimpleNamespace(track_combo=FakeCombo())
        dock._available_recording_tracks = lambda: [2000000, 1000000]

        with patch.object(helper, "get_app", return_value=fake_app):
            tracks = helper.AudioRecordingDockContent._recording_track_stack(dock, 3)

        self.assertEqual(tracks, [2000000, 1000000, 500000])
        self.assertEqual(created, [500000])

    def test_no_track_disables_preview_and_timeline_insertion(self):
        helper = self.audio_recording_module

        combo = types.SimpleNamespace(currentData=lambda: helper.NO_RECORDING_TRACK)
        dock = types.SimpleNamespace(track_combo=combo)
        dock._recording_timeline_enabled = lambda: (
            helper.AudioRecordingDockContent._recording_timeline_enabled(dock)
        )

        previews = helper.AudioRecordingDockContent._recording_preview_payloads(
            dock, 0.0, 1.0, [])
        assignments = helper.AudioRecordingDockContent._recording_track_assignments(
            dock, ["mic", "screen", "webcam"])

        self.assertEqual(previews, [])
        self.assertEqual(assignments, {})
        self.assertFalse(dock._recording_timeline_enabled())

    def test_recording_import_without_track_only_adds_project_file(self):
        helper = self.audio_recording_module
        files_model = types.SimpleNamespace(add_files=MagicMock())
        refresh_signal = types.SimpleNamespace(emit=MagicMock())
        dock = types.SimpleNamespace(
            window=types.SimpleNamespace(
                files_model=files_model,
                refreshFilesSignal=refresh_signal,
            ),
            _add_recording_to_timeline=MagicMock(),
        )

        result = helper.AudioRecordingDockContent._import_recording(
            dock, "Screen-1.mp4", "screen", add_to_timeline=False)

        self.assertIsNone(result)
        files_model.add_files.assert_called_once_with(
            "Screen-1.mp4",
            quiet=True,
            prevent_image_seq=True,
            prevent_recent_folder=True,
        )
        refresh_signal.emit.assert_called_once_with()
        dock._add_recording_to_timeline.assert_not_called()

    def test_screen_source_maps_to_backend_value(self):
        helper = self.audio_recording_module

        class FakeScreenCombo:
            def __init__(self, source):
                self.source = source

            def currentData(self):
                return self.source

        dock = types.SimpleNamespace(screen_display_edit=FakeScreenCombo({"display": ":0.0"}))
        dock._selected_screen_source = lambda: helper.AudioRecordingDockContent._selected_screen_source(dock)
        self.assertEqual(
            helper.AudioRecordingDockContent._screen_display_value(dock, ":1.0"),
            ":0.0",
        )

        dock.screen_display_edit.source = {}
        self.assertEqual(
            helper.AudioRecordingDockContent._screen_display_value(dock, ":1.0"),
            ":1.0",
        )

    def test_selected_screen_source_sets_full_screen_bounds(self):
        helper = self.audio_recording_module

        class FakeScreenCombo:
            def currentData(self):
                return {
                    "id": "screen-2",
                    "display": "desktop",
                    "x": 1920,
                    "y": 0,
                    "width": 1280,
                    "height": 720,
                    "all": False,
                }

        class FakeSpin:
            def __init__(self):
                self.value = None

            def setValue(self, value):
                self.value = value

        class FakeLabel:
            def __init__(self):
                self.text = ""

            def setText(self, text):
                self.text = text

        dock = types.SimpleNamespace(
            screen_display_edit=FakeScreenCombo(),
            screen_x_spin=FakeSpin(),
            screen_y_spin=FakeSpin(),
            screen_width_spin=FakeSpin(),
            screen_height_spin=FakeSpin(),
            screen_status_label=FakeLabel(),
        )
        dock._selected_screen_source = lambda: helper.AudioRecordingDockContent._selected_screen_source(dock)

        helper.AudioRecordingDockContent._set_screen_to_selected_source(dock)

        self.assertEqual(dock.screen_x_spin.value, 1920)
        self.assertEqual(dock.screen_y_spin.value, 0)
        self.assertEqual(dock.screen_width_spin.value, 1280)
        self.assertEqual(dock.screen_height_spin.value, 720)
        self.assertIn("1280x720", dock.screen_status_label.text)

    def test_windows_region_uses_native_cursor_pixels_and_monitor_bounds(self):
        helper = self.recording_widgets_module
        overlay = helper.WindowsRegionSelectorOverlay(
            capture_geometry=(3840, 0, 3840, 2160),
            overlay_geometry=QRect(2560, 0, 2560, 1440),
        )

        class FakeMouseEvent:
            def buttons(self):
                return Qt.LeftButton

        positions = [(3990, 150), (4590, 750), (8000, 3000)]
        with patch.object(helper, "_windows_cursor_pos", side_effect=positions):
            overlay.mousePressEvent(FakeMouseEvent())
            overlay.mouseMoveEvent(FakeMouseEvent())
            overlay.mouseReleaseEvent(FakeMouseEvent())

        # The returned rectangle is in native gdigrab pixels, not the
        # 150%-scaled Qt coordinate space, and it cannot escape Screen 2.
        self.assertEqual(overlay.selected_geometry(), (3990, 150, 3690, 2010))

    def test_windows_region_overlay_matches_high_dpi_qt_screen(self):
        helper = self.recording_widgets_module
        fake_screen = types.SimpleNamespace(
            geometry=lambda: QRect(2560, 0, 2560, 1440),
            devicePixelRatio=lambda: 1.5,
        )
        with patch.object(helper.QApplication, "screens", return_value=[fake_screen]):
            geometry = helper._windows_region_overlay_geometry((3840, 0, 3840, 2160))

        self.assertEqual(geometry, QRect(2560, 0, 2560, 1440))

    def test_windows_physical_point_maps_with_each_monitors_dpi(self):
        helper = self.recording_widgets_module
        monitors = [
            {"x": 0, "y": 0, "width": 3840, "height": 2160},
            {"x": 3840, "y": 0, "width": 1920, "height": 1080},
        ]
        screens = [
            types.SimpleNamespace(
                geometry=lambda: QRect(0, 0, 2560, 1440),
                devicePixelRatio=lambda: 1.5),
            types.SimpleNamespace(
                geometry=lambda: QRect(2560, 0, 1920, 1080),
                devicePixelRatio=lambda: 1.0),
        ]
        with patch.object(helper, "windows_monitor_geometries", return_value=monitors), \
                patch.object(helper.QApplication, "screens", return_value=screens):
            high_dpi_point = helper._windows_physical_point_to_qt(1920, 1080)
            standard_dpi_point = helper._windows_physical_point_to_qt(4800, 540)

        self.assertEqual(high_dpi_point, QPoint(1280, 720))
        self.assertEqual(standard_dpi_point, QPoint(3520, 540))

    def test_windows_window_picker_excludes_overlay_not_openshot_window(self):
        helper = self.recording_widgets_module

        class FakePoint:
            def __init__(self, x, y):
                self.x = x
                self.y = y

        class FakeUser32:
            def IsWindowVisible(self, _hwnd):
                return True

            def IsIconic(self, _hwnd):
                return False

            def GetWindowThreadProcessId(self, *_args):
                raise AssertionError("The picker must not reject all OpenShot windows")

            def EnumWindows(self, callback, _lparam):
                for hwnd in (11, 22, 33):
                    if not callback(hwnd, 0):
                        break
                return True

        rects = {
            11: types.SimpleNamespace(left=0, top=0, right=4000, bottom=2200),
            22: types.SimpleNamespace(left=600, top=300, right=3200, bottom=1900),
            33: types.SimpleNamespace(left=0, top=0, right=4000, bottom=2200),
        }
        fake_types = types.SimpleNamespace(POINT=FakePoint)
        with patch.object(helper, "_windows_types", return_value=fake_types), \
                patch.object(helper, "_windows_user32", return_value=FakeUser32()), \
                patch.object(helper, "_windows_enum_proc_type", return_value=lambda callback: callback), \
                patch.object(helper, "_windows_window_title", side_effect=lambda hwnd: "OpenShot" if hwnd == 22 else "Other"), \
                patch.object(helper, "_windows_window_rect", side_effect=lambda hwnd: rects[hwnd]):
            picked = helper._windows_pick_window_at(1000, 800, excluded_hwnd=11)

        self.assertEqual(picked[0], 22)
        self.assertEqual(picked[2], "OpenShot")

    def test_windows_all_screens_region_uses_qt_virtual_desktop_overlay(self):
        helper = self.recording_widgets_module
        virtual_geometry = QRect(-1280, 0, 3200, 1080)
        fake_primary = types.SimpleNamespace(virtualGeometry=lambda: virtual_geometry)
        fake_overlay = MagicMock()
        fake_overlay.select.return_value = (10, 20, 30, 40)
        with patch.object(helper, "windows_virtual_screen_geometry", return_value=(-1920, 0, 3840, 1080)), \
                patch.object(helper.QApplication, "primaryScreen", return_value=fake_primary), \
                patch.object(helper, "WindowsRegionSelectorOverlay", return_value=fake_overlay) as overlay_class:
            result = helper.pick_windows_region()

        self.assertEqual(result, (10, 20, 30, 40))
        overlay_class.assert_called_once_with(
            None, (-1920, 0, 3840, 1080), virtual_geometry)

    def test_select_region_limits_windows_picker_to_selected_screen(self):
        helper = self.audio_recording_module
        button = lambda: types.SimpleNamespace(setChecked=MagicMock())
        dock = types.SimpleNamespace(
            region_button=button(),
            full_screen_button=button(),
            window_button=button(),
            _screen_window_id="old-window",
            _set_hide_openshot_default=MagicMock(),
            _hide_openshot_for_picker=MagicMock(return_value=None),
            _restore_openshot_window=MagicMock(),
            _selected_screen_source=MagicMock(return_value={
                "id": "screen-2", "x": 3840, "y": 0,
                "width": 3840, "height": 2160, "all": False,
            }),
            _set_screen_target=MagicMock(),
            screen_status_label=types.SimpleNamespace(setText=MagicMock()),
        )
        with patch.object(helper, "screen_capture_backend_is_wayland", return_value=False), \
                patch.object(helper.sys, "platform", "win32"), \
                patch.object(helper.QApplication, "processEvents"), \
                patch.object(helper, "pick_screen_region", return_value=(4000, 100, 800, 600)) as picker:
            helper.AudioRecordingDockContent._select_region(dock)

        picker.assert_called_once_with(
            dock,
            capture_geometry=(3840, 0, 3840, 2160),
        )
        dock._set_screen_target.assert_called_once()

    def test_region_and_window_overlays_cancel_escape_from_application(self):
        helper = self.recording_widgets_module

        class EscapeEvent:
            def type(self):
                return QEvent.KeyPress

            def key(self):
                return Qt.Key_Escape

        with patch.object(helper, "windows_virtual_screen_geometry", return_value=(0, 0, 1920, 1080)):
            overlays = (
                helper.RegionSelectorOverlay(),
                helper.WindowsWindowSelectorOverlay(),
            )
        for overlay in overlays:
            overlay.reject = MagicMock()
            self.assertTrue(overlay.eventFilter(self.app, EscapeEvent()))
            overlay.reject.assert_called_once_with()

    def test_windows_screen_job_uses_selected_monitor_bounds(self):
        helper = self.audio_recording_module

        class FakeCard:
            def __init__(self, checked):
                self.checked = checked

            def isChecked(self):
                return self.checked

        class FakeControl:
            def __init__(self, value):
                self.current = value

            def currentData(self):
                return self.current

            def value(self):
                return self.current

            def setValue(self, value):
                self.current = value

            def isChecked(self):
                return bool(self.current)

        class FakeSettings:
            pass

        captured = []

        class FakeReader:
            def __init__(self, settings):
                captured.append(settings)

        for x, width, all_screens in (
                (1920, 1920, False),
                (-1920, 1920, False),
                (-1920, 5760, True)):
            source = {
                "id": "screen-2",
                "display": "desktop",
                "x": x,
                "y": 0,
                "width": width,
                "height": 1080,
                "all": all_screens,
            }
            dock = types.SimpleNamespace(
                screen_card=FakeCard(True),
                camera_card=FakeCard(False),
                video_fps_combo=FakeControl(30),
                screen_x_spin=FakeControl(x),
                screen_y_spin=FakeControl(0),
                screen_width_spin=FakeControl(width),
                screen_height_spin=FakeControl(1080),
                full_screen_button=FakeControl(True),
                window_button=FakeControl(False),
                region_button=FakeControl(False),
                capture_cursor_combo=FakeControl(True),
                system_audio_combo=FakeControl(False),
                _screen_window_id="",
                _recording_preview_file_ids={},
                _selected_screen_source=lambda source=source: source,
                _system_audio_available=lambda: False,
                _safe_even_dimension=helper.AudioRecordingDockContent._safe_even_dimension,
                _screen_recording_bit_rate=lambda width, height, fps: 8000000,
                _next_named_recording_path=lambda prefix, extension: "Screen.mp4",
            )

            with patch.object(helper, "screen_capture_backend", return_value=7), \
                    patch.object(helper, "screen_capture_backend_is_wayland", return_value=False), \
                    patch.object(helper, "screen_capture_backend_is_windows", return_value=True), \
                    patch.object(helper, "screen_capture_backend_is_mac", return_value=False), \
                    patch.object(helper.openshot, "ScreenCaptureSettings", FakeSettings), \
                    patch.object(helper.openshot, "ScreenCaptureReader", FakeReader):
                jobs = helper.AudioRecordingDockContent._build_video_jobs(dock)

            self.assertEqual(len(jobs), 1)
            settings = captured[-1]
            self.assertEqual((settings.x, settings.y), (x, 0))
            self.assertEqual((settings.width, settings.height), (width, 1080))
            self.assertEqual(settings.display, "desktop")

    def test_timeline_recording_previews_build_audio_and_video_clip_data(self):
        timeline_module = self.timeline_module
        helper = types.SimpleNamespace(
            _recording_preview_clips=[],
            fps_float=30.0,
            geometry=types.SimpleNamespace(mark_dirty=lambda: setattr(helper, "dirty", True)),
            dirty=False,
            updated=False,
        )
        helper.update = lambda: setattr(helper, "updated", True)

        timeline_module.TimelineView.set_audio_recording_previews(helper, [
            {
                "id": "preview-mic",
                "source_type": "mic",
                "position": 12.0,
                "track": 2,
                "duration": 3.0,
                "audio_data": [0.1, 0.4],
                "audio_data_rms": [0.05, 0.2],
                "audio_data_rate": 200,
            },
            {
                "id": "preview-screen",
                "source_type": "screen",
                "position": 12.0,
                "track": 3,
                "duration": 3.0,
                "file_id": "temp-screen",
                "width": 1920,
                "height": 1080,
                "fps": 30.0,
            },
        ])

        self.assertTrue(helper.dirty)
        self.assertTrue(helper.updated)
        self.assertEqual(len(helper._recording_preview_clips), 2)

        mic_data = helper._recording_preview_clips[0].data
        self.assertEqual(mic_data["layer"], 2)
        self.assertTrue(mic_data["reader"]["has_audio"])
        self.assertFalse(mic_data["reader"]["has_video"])
        self.assertEqual(mic_data["ui"]["audio_data"], [0.1, 0.4])
        self.assertEqual(mic_data["ui"]["audio_data_rms"], [0.05, 0.2])
        self.assertEqual(mic_data["ui"]["audio_data_rate"], 200)
        self.assertEqual(mic_data["ui"]["audio_data_format"], "absolute_peak_v2")

        screen_data = helper._recording_preview_clips[1].data
        self.assertEqual(screen_data["file_id"], "temp-screen")
        self.assertEqual(screen_data["reader"]["id"], "temp-screen")
        self.assertTrue(screen_data["reader"]["has_video"])
        self.assertEqual(screen_data["reader"]["width"], 1920)
        self.assertEqual(screen_data["reader"]["video_length"], 90)


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for Film Grain preset helpers
"""

import os
import sys
import unittest

PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.film_grain_presets import (  # noqa: E402
    FILM_GRAIN_PRESET_16MM_CLASSIC,
    FILM_GRAIN_PRESET_35MM_CLASSIC,
    FILM_GRAIN_PRESET_35MM_FINE,
    FILM_GRAIN_PRESET_35MM_GRITTY,
    FILM_GRAIN_PRESET_HIGH_ISO,
    FILM_GRAIN_PRESET_SUPER_8,
    apply_film_grain_preset,
    is_film_grain_effect,
)


class FilmGrainPresetTests(unittest.TestCase):
    def test_presets_initialize_all_visible_keyframe_controls(self):
        for preset in (
            FILM_GRAIN_PRESET_35MM_FINE,
            FILM_GRAIN_PRESET_35MM_CLASSIC,
            FILM_GRAIN_PRESET_35MM_GRITTY,
            FILM_GRAIN_PRESET_16MM_CLASSIC,
            FILM_GRAIN_PRESET_SUPER_8,
            FILM_GRAIN_PRESET_HIGH_ISO,
        ):
            payload = apply_film_grain_preset({"class_name": "FilmGrain"}, preset)
            for key in (
                "amount",
                "size",
                "softness",
                "clump",
                "shadows",
                "midtones",
                "highlights",
                "color_amount",
                "color_variation",
                "evolution",
                "coherence",
            ):
                self.assertIn("Points", payload[key])
                self.assertEqual(payload[key]["Points"][0]["co"]["X"], 1.0)
            self.assertEqual(payload["seed"], 1)

    def test_presets_get_progressively_stronger_and_coarser(self):
        fine = apply_film_grain_preset({}, FILM_GRAIN_PRESET_35MM_FINE)
        classic = apply_film_grain_preset({}, FILM_GRAIN_PRESET_35MM_CLASSIC)
        gritty = apply_film_grain_preset({}, FILM_GRAIN_PRESET_35MM_GRITTY)
        sixteen = apply_film_grain_preset({}, FILM_GRAIN_PRESET_16MM_CLASSIC)
        super8 = apply_film_grain_preset({}, FILM_GRAIN_PRESET_SUPER_8)

        self.assertLess(fine["amount"]["Points"][0]["co"]["Y"], classic["amount"]["Points"][0]["co"]["Y"])
        self.assertLess(classic["amount"]["Points"][0]["co"]["Y"], gritty["amount"]["Points"][0]["co"]["Y"])
        self.assertLess(gritty["amount"]["Points"][0]["co"]["Y"], sixteen["amount"]["Points"][0]["co"]["Y"])
        self.assertLess(sixteen["amount"]["Points"][0]["co"]["Y"], super8["amount"]["Points"][0]["co"]["Y"])
        self.assertLess(fine["size"]["Points"][0]["co"]["Y"], super8["size"]["Points"][0]["co"]["Y"])
        self.assertLess(fine["clump"]["Points"][0]["co"]["Y"], super8["clump"]["Points"][0]["co"]["Y"])

    def test_seed_is_plain_int_and_preserved_when_existing(self):
        payload = apply_film_grain_preset({"class_name": "FilmGrain", "seed": 1234}, FILM_GRAIN_PRESET_16MM_CLASSIC)
        self.assertEqual(payload["seed"], 1234)
        self.assertIsInstance(payload["seed"], int)

    def test_is_film_grain_effect(self):
        self.assertTrue(is_film_grain_effect({"class_name": "FilmGrain"}))
        self.assertFalse(is_film_grain_effect({"class_name": "ColorGrade"}))
        self.assertFalse(is_film_grain_effect(None))


if __name__ == "__main__":
    unittest.main()



"""
 @file
 @brief Unit tests for project keyframe frame-number scaling
 @author Jonathan Thomas <jonathan@openshot.org>

 @section LICENSE

 Copyright (c) 2008-2026 OpenShot Studios, LLC
 (http://www.openshotstudios.com). This file is part of
 OpenShot Video Editor (http://www.openshot.org), an open-source project
 dedicated to delivering high quality video editing and animation solutions
 to the world.

 OpenShot Video Editor is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 OpenShot Video Editor is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with OpenShot Library.  If not, see <http://www.gnu.org/licenses/>.
 """

import os
import sys
import unittest


PATH = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
if PATH not in sys.path:
    sys.path.append(PATH)

from classes.keyframe_scaler import KeyframeScaler  # noqa: E402


def keyframe(*frames):
    return {
        "Points": [
            {"co": {"X": float(frame), "Y": float(index)}, "interpolation": 0}
            for index, frame in enumerate(frames)
        ]
    }


def frame_numbers(data):
    return [point["co"]["X"] for point in data["Points"]]


class KeyframeScalerTests(unittest.TestCase):
    def test_scales_clip_effect_transition_colorgrade_and_time_keyframes(self):
        data = {
            "clips": [{
                "id": "clip-1",
                "alpha": keyframe(1, 30),
                "location": {
                    "red": keyframe(1, 30),
                    "green": keyframe(1, 30),
                    "blue": keyframe(1, 30),
                    "alpha": keyframe(1, 30),
                },
                "time": {
                    "Points": [
                        {"co": {"X": 1.0, "Y": 1.0}, "interpolation": 0},
                        {"co": {"X": 30.0, "Y": 30.0}, "interpolation": 0},
                    ]
                },
                "effects": [{
                    "class_name": "ColorGrade",
                    "wheels": {
                        "enabled_keyframes": keyframe(1, 30),
                        "global": {
                            "color_keyframes": {
                                "red": keyframe(1, 30),
                                "green": keyframe(1, 30),
                                "blue": keyframe(1, 30),
                                "alpha": keyframe(1, 30),
                            },
                            "amount_keyframes": keyframe(1, 30),
                            "luma_keyframes": keyframe(1, 30),
                        },
                    },
                    "curve": {
                        "enabled": keyframe(1, 30),
                        "nodes": [{
                            "id": 1,
                            "x": keyframe(1, 30),
                            "y": {
                                "Points": [
                                    {"co": {"X": 1.0, "Y": 0.2}, "interpolation": 0},
                                    {"co": {"X": 30.0, "Y": 0.8}, "interpolation": 0},
                                ]
                            },
                            "left_handle_x": keyframe(1, 30),
                            "right_handle_y": keyframe(1, 30),
                        }],
                    },
                }],
            }],
            "effects": [{
                "id": "transition-1",
                "brightness": keyframe(1, 30),
            }],
        }

        KeyframeScaler(2.0)(data)

        clip = data["clips"][0]
        self.assertEqual(frame_numbers(clip["alpha"]), [1.0, 60])
        self.assertEqual(frame_numbers(clip["location"]["red"]), [1.0, 60])
        self.assertEqual(frame_numbers(clip["time"]), [1.0, 60])
        self.assertEqual([point["co"]["Y"] for point in clip["time"]["Points"]], [1.0, 60])

        effect = clip["effects"][0]
        wheels = effect["wheels"]
        self.assertEqual(frame_numbers(wheels["enabled_keyframes"]), [1.0, 60])
        self.assertEqual(frame_numbers(wheels["global"]["color_keyframes"]["red"]), [1.0, 60])
        self.assertEqual(frame_numbers(wheels["global"]["amount_keyframes"]), [1.0, 60])
        self.assertEqual(frame_numbers(wheels["global"]["luma_keyframes"]), [1.0, 60])

        curve = effect["curve"]
        node = curve["nodes"][0]
        self.assertEqual(frame_numbers(curve["enabled"]), [1.0, 60])
        self.assertEqual(frame_numbers(node["x"]), [1.0, 60])
        self.assertEqual(frame_numbers(node["y"]), [1.0, 60])
        self.assertEqual([point["co"]["Y"] for point in node["y"]["Points"]], [0.2, 0.8])
        self.assertEqual(frame_numbers(node["left_handle_x"]), [1.0, 60])
        self.assertEqual(frame_numbers(node["right_handle_y"]), [1.0, 60])

        self.assertEqual(frame_numbers(data["effects"][0]["brightness"]), [1.0, 60])

    def test_scales_nested_colorgrade_keyframes_without_color_channel_shortcut(self):
        data = {
            "clips": [{
                "effects": [{
                    "wheels": {
                        "highlights": {
                            "color_keyframes": {
                                "red": keyframe(1, 30),
                                "green": keyframe(1, 30),
                                "blue": keyframe(1, 30),
                                "alpha": keyframe(1, 30),
                            },
                        },
                    },
                    "curve": {
                        "nodes": [
                            {"id": 0, "right_handle_x": keyframe(1, 30)},
                            {"id": 1, "left_handle_y": keyframe(1, 30)},
                        ],
                    },
                }],
            }],
        }

        KeyframeScaler(2.0)(data)

        effect = data["clips"][0]["effects"][0]
        color = effect["wheels"]["highlights"]["color_keyframes"]
        self.assertEqual(frame_numbers(color["red"]), [1.0, 60])
        self.assertEqual(frame_numbers(color["green"]), [1.0, 60])
        self.assertEqual(frame_numbers(color["blue"]), [1.0, 60])
        self.assertEqual(frame_numbers(color["alpha"]), [1.0, 60])
        self.assertEqual(frame_numbers(effect["curve"]["nodes"][0]["right_handle_x"]), [1.0, 60])
        self.assertEqual(frame_numbers(effect["curve"]["nodes"][1]["left_handle_y"]), [1.0, 60])


if __name__ == "__main__":
    unittest.main()



"""Tests for main-window startup state handling."""

import unittest
from unittest.mock import patch

from qt_api import Qt

from classes.app import OpenShotApp


class FakeWindow:
    def __init__(self, state):
        self.state = state
        self.calls = []

    def windowState(self):
        return self.state

    def show(self):
        self.calls.append("show")

    def showMaximized(self):
        self.calls.append("showMaximized")

    def showFullScreen(self):
        self.calls.append("showFullScreen")

    def showNormal(self):
        self.calls.append("showNormal")

    def _restore_saved_window_state(self):
        self.calls.append("restoreState")


class AppWindowStateTests(unittest.TestCase):
    def test_restored_maximized_window_is_shown_maximized(self):
        window = FakeWindow(Qt.WindowMaximized)

        with patch(
                "classes.app.QTimer.singleShot",
                lambda _delay, callback: callback()):
            OpenShotApp._show_main_window(window)

        self.assertEqual(
            window.calls,
            ["show", "showNormal", "showMaximized", "restoreState"])

    def test_restored_fullscreen_window_is_shown_fullscreen(self):
        window = FakeWindow(Qt.WindowFullScreen)

        with patch(
                "classes.app.QTimer.singleShot",
                lambda _delay, callback: callback()):
            OpenShotApp._show_main_window(window)

        self.assertEqual(
            window.calls,
            ["show", "showNormal", "showFullScreen", "restoreState"])

    def test_normal_window_uses_plain_show(self):
        window = FakeWindow(Qt.WindowNoState)

        with patch(
                "classes.app.QTimer.singleShot",
                lambda _delay, callback: callback()):
            OpenShotApp._show_main_window(window)

        self.assertEqual(window.calls, ["show", "restoreState"])


if __name__ == "__main__":
    unittest.main()
